import requests
import csv
import sys
import os
from io import StringIO

PERPLEXITY_API_KEY = os.getenv('PERPLEXITY_API_KEY', )


def find_between( s, first, last ):
    try:
        start = s.index( first ) + len( first )
        end = s.index( last, start )
        return s[start:end]
    except ValueError:
        return ""

def get_fields_info(fields, language='English'):
    """Returns information about data fields"""
    url = "https://api.perplexity.ai/chat/completions"
    headers = {"Authorization": f"Bearer {PERPLEXITY_API_KEY}"}
    payload = {
        "model": "sonar",
        "messages": [
            {"role": "system", "content": "Be precise and concise, provide data output only CSV or JSON, accrording to request"},
            {"role": "user", "content": (
                f"Please describe in {language} these fields delimited by comma: {fields}"
                "Please output as single csv table only with following fields: name and description"
            )},
        ],
        "response_format": {
                "type": "text",
        },
    }
    response = requests.post(url, headers=headers, json=payload).json()
    text = response["choices"][0]["message"]["content"]
    a_text = find_between(text, "```csv", "```").strip()
    if len(a_text) == 0:
        a_text = find_between(text, "```", "```").strip()
    f = StringIO()
    f.write(a_text)
    f.seek(0)
    table = {}
    dr = csv.reader(f, delimiter=',')    
    n = 0
    for r in dr:
        n += 1
        if n == 1: continue
        table[r[0]] = r[1]
    return table



def get_description(data, language='English'):
    url = "https://api.perplexity.ai/chat/completions"
    headers = {"Authorization": f"Bearer {PERPLEXITY_API_KEY}"}
    payload = {
        "model": "sonar",
        "messages": [
            {"role": "system", "content": "Be precise and concise, provide data output only CSV or JSON, accrording to request"},
            {"role": "user", "content": (
                f"""
I have the following CSV data: 
{data}
Please provide short description in {language} about this data in English. Consider this data as sample of the bigger dataset.Don't generate any code and data examples""")},
        ],
        "response_format": {
                "type": "text",
        },
    }
    response = requests.post(url, headers=headers, json=payload).json()
    return response["choices"][0]["message"]["content"]




if __name__ == "__main__":
    print(get_fields_info(sys.argv[1], sys.argv[2]))


# -*- coding: utf8 -*-
DATE_PATTERNS = ["%d.%m.%Y", "%Y-%m-%d", "%y-%m-%d", "%Y-%m-%dT%H:%M:%S",
                 "%Y-%m-%d %H:%M:%S",
                 "%d.%m.%Y %H:%M"]
DEFAULT_DICT_SHARE = 70



SUPPORTED_FILE_TYPES = ['xls', 'xlsx', 'csv', 'xml', 'json', 'jsonl', 'yaml', 'tsv', 'sql', 'bson', 'parquet', 'orc', 'avro']
COMPRESSED_FILE_TYPES = ['gz', 'xz', 'zip', 'lz4', '7z', 'bz2']
BINARY_FILE_TYPES = ['xls', 'xlsx', 'bson', 'parquet', 'irc'] + COMPRESSED_FILE_TYPES

DEFAULT_OPTIONS = {'encoding' : 'utf8',
                   'delimiter' : ',',
                   'limit' : 1000
                   }



# -*- coding: utf-8 -*-
KPP_CODE = {
    u'01' : u'Постановка на учет в налоговом органе российской организации в качестве налогоплательщика по месту ее нахождения',
    u'02' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения ее обособленного подразделения в зависимости от вида подразделения',
    u'03' : u'Постановка на учет налогоплательщика – российской организации по месту нахождения ее филиала, не исполняющего обязанности организации по уплате налогов и сборов',
    u'04' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения ее обособленного подразделения в зависимости от вида подразделения',
    u'05' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения ее обособленного подразделения в зависимости от вида подразделения',
    u'06' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения принадлежащего ей недвижимого имущества (за исключением транспортных средств) - в зависимости от вида имущества',
    u'07' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения принадлежащего ей недвижимого имущества (за исключением транспортных средств) - в зависимости от вида имущества',
    u'08' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения принадлежащего ей недвижимого имущества (за исключением транспортных средств) - в зависимости от вида имущества',
    u'30' : u'Российская организация - налоговый агент, не учтенная в качестве налогоплательщика',
    u'31' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения обособленного подразделения, в отношении которого не проведена процедура оформления в соответствии с пунктом 3 статьи 55 Гражданского кодекса Российской Федерации, исполняющего обязанности организации по уплате налогов и сборов',
    u'32' : u'Постановка на учет налогоплательщика - российской организации по месту нахождения обособленного подразделения, в отношении которого не проведена процедура оформления в соответствии с пунктом 3 статьи 55 Гражданского кодекса Российской Федерации, не исполняющего обязанности организации по уплате налогов и сборов',
    u'51' : u'Постановка на учет отделений иностранных организаций',
    u'52' : u'Постановка на учет отделений иностранных организаций в Российской Федерации, созданных филиалом этой иностранной организации в иностранном государстве',
    u'53' : u'резервный код',
    u'54' : u'резервный код',
    u'55' : u'резервный код',
    u'56' : u'резервный код',
    u'57' : u'резервный код',
    u'58' : u'резервный код',
    u'59' : u'резервный код',
    u'60' : u'Постановка на учет посольств иностранных государств',
    u'61' : u'Постановка на учет консульств иностранных государств',
    u'62' : u'Постановка на учет представительств, приравненных к дипломатическим',
    u'63' : u'Постановка на учет международных организаций',
    u'64' : u'резервный код',
    u'65' : u'резервный код',
    u'66' : u'резервный код',
    u'67' : u'резервный код',
    u'68' : u'резервный код',
    u'69' : u'резервный код',
    u'70' : u'Постановка на учет иностранных и международных организаций, имеющих недвижимое имущество в Российской Федерации, за исключением транспортных средств, относящихся к недвижимому имуществу',
    u'71' : u'Постановка на учет иностранных и международных организаций, имеющих транспортные средства в Российской Федерации, не относящиеся к недвижимому имуществу',
    u'72' : u'Постановка на учет иностранных и международных организаций, имеющих морские транспортные средства в Российской Федерации',
    u'73' : u'Постановка на учет иностранных и международных организаций, имеющих речные транспортные средства в Российской Федерации',
    u'74' : u'Постановка на учет иностранных и международных организаций, имеющих воздушные транспортные средства в Российской Федерации',
    u'75' : u'Постановка на учет иностранных и международных организаций, имеющих космические объекты в Российской Федерации',
    u'76' : u'резервный код',
    u'77' : u'резервный код',
    u'78' : u'резервный код',
    u'79' : u'резервный код',
    u'80' : u'Учет иностранных и международных организаций в связи с открытием в банках рублевых счетов типа "Т" (текущие)',
    u'81' : u'Учет иностранных и международных организаций в связи с открытием счетов в банках типа "И" (инвестиционные)',
    u'82' : u'учет иностранных и международных организаций в связи с открытием счетов в банках типа "С" (специальные)',
    u'83' : u'Учет иностранных и международных организаций в связи с открытием в банках счетов типа "Т" (текущие) в иностранной валюте',
    u'84' : u'Учет иностранных и международных организаций в связи с открытием корреспондентских счетов в банка',
    }


OGRN_CODES  = {
    u'1' : u'юридическое лицо',
    u'5' : u'юридическое лицо',
    u'3' : u'индивидуальный предприниматель',
    }


SOUN_SCHEMA = {'1' : {'name' : 'kod', 'type' : 'string'},
               '2' : {'name' : 'vid', 'type' : 'string'},
               '3' : {'name' : 'kodp', 'type' : 'string'},
               '4' : {'name' : 'kodv', 'type' : 'string'},
               '5' : {'name' : 'naimk', 'type' : 'string'},
               '6' : {'name' : 'naim', 'type' : 'string'},
               '7' : {'name' : 'psono', 'type' : 'int'},
               '8' : {'name' : 'puch', 'type' : 'int'},
               '9' : {'name' : 'potchdok', 'type' : 'int'},
               '10' : {'name' : 'potch', 'type' : 'int'},
               '11' : {'name' : 'spro_u', 'type' : 'string'},
               '12' : {'name' : 'spro_f', 'type' : 'string'},
               '13' : {'name' : 'inn', 'type' : 'string'},
               '14' : {'name' : 'kpp', 'type' : 'string'},
               '15' : {'name' : 'adres', 'type' : 'string'},
               '16' : {'name' : 'tel', 'type' : 'string'},
               '17' : {'name' : 'email', 'type' : 'string'},
               '18' : {'name' : 'cite', 'type' : 'string'},
               '19' : {'name' : 'dokum', 'type' : 'string'},
               '20' : {'name' : 'nomdok', 'type' : 'string'},
               '21' : {'name' : 'datadok', 'type' : 'date'},
               '22' : {'name' : 'datan', 'type' : 'date'},
               '23' : {'name' : 'datak', 'type' : 'date'},
               '24' : {'name' : 'coment', 'type' : 'string'},
               }

SSRF_SCHEMA = {'1' : {'name' : 'code', 'type' : 'string'},
               '2' : {'name' : 'name', 'type' : 'string'},
               }

OKV_SCHEMA = {'1' : {'name' : 'id', 'type' : 'int'},
              '2' : {'name' : 'ncode', 'type' : 'string'},
              '3' : {'name' : 'ccode', 'type' : 'string'},
              '4' : {'name' : 'name', 'type' : 'string'},
              '5' : {'name' : 'countries', 'type' : 'string'},
              }

PLANSCHET_SCHEMA = {'1' : {'name' : 'id', 'type' : 'int'},
                    '2' : {'name' : 'code', 'type' : 'string'},
                    '3' : {'name' : 'name', 'type' : 'string'},
                    '4' : {'name' : 'stype', 'type' : 'int'},
                    '5' : {'name' : 'parent_id', 'type' : 'int'},
                    '6' : {'name' : 'level', 'type' : 'int'},
                    }


RUS_INN_FACTOR = (2, 4, 10, 3, 5, 9, 4, 6, 8)

RUS_ACCOUNT_MASK = [7,1,3,7,1,3,7,1,3,7,1,3,7,1,3,7,1,3,7,1,3,7,1]

def calc_personal_check_digits(number):
    """Calculate the check digits for the 12-digit personal ИНН."""
    weights = (7, 2, 4, 10, 3, 5, 9, 4, 6, 8)
    d1 = str(sum(w * int(n) for w, n in zip(weights, number)) % 11 % 10)
    weights = (3, 7, 2, 4, 10, 3, 5, 9, 4, 6, 8)
    d2 = str(sum(w * int(n) for w, n in zip(weights, number[:10] + d1)) % 11 % 10)
    return d1 + d2


def _check_inn(code):
    """Validates INN code"""
    if code is not None and code.isdigit():
        if len(code) == 10:
            asum = 0
            for i in range(0, 9):
                asum += int(code[i]) * RUS_INN_FACTOR[i]
            asum = (asum % 11) % 10
            return (int(code[9]) == asum)
        elif len(code) == 12:
            weights = (7, 2, 4, 10, 3, 5, 9, 4, 6, 8)
            d1 = str(sum(w * int(n) for w, n in zip(weights, code)) % 11 % 10)
            weights = (3, 7, 2, 4, 10, 3, 5, 9, 4, 6, 8)
            d2 = str(sum(w * int(n) for w, n in zip(weights, code[:10] + d1)) % 11 % 10)
            chsum = d1 + d2
            return chsum == code[-2:]
    return False


def _check_ogrn(ogrn):
    """Validates OGRN code"""
    if not ogrn:
        return False
    if len(ogrn) == 13:
        delimeter = 11
    elif len(ogrn) == 15:
        delimeter = 13
    else:
        return False
    main_part = int(ogrn[:-1]) % delimeter % 10
    checksum = int(ogrn[-1])

    return main_part == checksum



# -*- coding: utf8 -*-
import duckdb
import logging
from iterable.helpers.detect import open_iterable
from tqdm import tqdm

from pymongo import MongoClient
from elasticsearch import Elasticsearch

ITERABLE_OPTIONS_KEYS = ['tagname', 'delimiter', 'encoding', 'start_line', 'page']

DUCKABLE_FILE_TYPES = ['parquet', 'csv', 'jsonl', 'json', 'jsonl.gz']
DUCKABLE_CODECS = ['gz', 'zst']


DEFAULT_BATCH_SIZE = 50000

def get_iterable_options(options):
    out = {}
    for k in ITERABLE_OPTIONS_KEYS:
        if k in options.keys():
            out[k] = options[k]
    return out            


class BasicIngester:
    def __init__(self):
        pass
    
    def ingest(self, batch):
        raise NotImplemented

class ElasticIngester(BasicIngester):
    def __init__(self, uri:str, api_key:str, search_index:str, document_id:str="id"):
        self.client = Elasticsearch(uri, api_key=api_key, verify_certs=False,ssl_show_warn=False, timeout=60, max_retries=10, retry_on_timeout=True)
        self._index = search_index
        self._item_id = document_id
        pass


    def ingest(self, batch):        
        documents = []
        for doc in batch:
            documents.append({ "index": { "_index": self._index, '_id' : doc[self._item_id]}})
            documents.append(doc)
        result = self.client.bulk(operations=documents, pipeline="ent-search-generic-ingestion")


class MongoIngester:
    def __init__(self, uri, db, table,do_drop=False):
        self.client = MongoClient(uri)
        self.db = self.client[db]
        if do_drop:
            del self.db[table]
        self.coll = self.db[table]


    def ingest(self, batch):
        result = self.coll.insert_many(batch)


class Ingester:
    def __init__(self, batch_size=DEFAULT_BATCH_SIZE):
        self.batch_size = batch_size
        pass

    def ingest(self, fromfiles, uri, db, table, options={}):
        for filename in fromfiles:
            self.ingest_single(filename, uri, db, table, options=options)

  
    def ingest_single(self, fromfile, uri, db, table, options={}):
        """Loads single file data contents to the schemaless database like MongoDB"""
        dbtype = options['dbtype']
        processor = None
        totals = -1
        skip = options['skip']
        use_totals = options['totals']if 'totals' in options.keys() else False
        do_drop = options['drop']if 'dro[]' in options.keys() else False

        if use_totals:
            parts = fromfile.rsplit('.', 2)
            if len(parts) == 2:
                if parts[-1].lower() in DUCKABLE_FILE_TYPES:
                    totals = duckdb.sql(f"select count(*) from '{fromfile}'").fetchone()[0]
            elif len(parts) == 3:
                if parts[-2].lower() in DUCKABLE_FILE_TYPES and parts[-1].lower() in DUCKABLE_CODECS:
                     totals = duckdb.sql(f"select count(*) from '{fromfile}'").fetchone()[0]
        if dbtype == 'mongodb':
            processor = MongoIngester(uri, db, table, do_drop=do_drop)
        elif dbtype == 'elastic':
            api_key = options['api_key']
            id_key = options['doc_id']
            processor = ElasticIngester(uri=uri, api_key=api_key, search_index=table, document_id=id_key)
        iterableargs = get_iterable_options(options)
        it_in = open_iterable(fromfile, mode='r', iterableargs=iterableargs)       
        logging.info(f'Ingesting data: filename {fromfile}, uri: {uri}, db {db}, table {table}')
        n = 0
        batch = []
        for row in tqdm(it_in, total=totals):
            n += 1 
            if skip is not None and skip > 0:
                if n < skip: continue
            batch.append(row)
            if n % self.batch_size == 0:
                processor.ingest(batch)
                batch = []
        if len(batch) > 0: 
            processor.ingest(batch)




# -*- coding: utf8 -*-
# FIXME: A lot of unoptimized code here, it could be better, shorter and some functions could be improved
import os
import io
import csv
import tempfile
from collections import OrderedDict
from bson import decode_file_iter
import json
import jsonlines
import xmltodict
import yaml
from pydantic import BaseModel
from typing import Optional
import duckdb
import pandas as pd
from iterable.helpers.detect import detect_file_type, TEXT_DATA_TYPES, detect_encoding_any

from openpyxl import load_workbook
import xlrd
from pyzstd import ZstdFile

from ..utils import get_file_type, detect_delimiter, detect_encoding, get_dict_value, get_dict_keys, _is_flat, buf_count_newlines_gen
from ..constants import SUPPORTED_FILE_TYPES
from ..formats.docx import analyze_docx
from ..ai.perplexity import get_fields_info, get_description

OBJECTS_ANALYZE_LIMIT = 10000


DUCKDB_TYPES = ['VARCHAR', 'DATE', 'JSON', 'BIGINT', 'DOUBLE', 'BOOLEAN']

def column_type_parse(column_type):
    is_array = (column_type[-2:] == '[]') 
    if is_array:
        text = column_type[:-2]
    else:
        text = column_type
    if text[:6] == 'STRUCT':        
        atype = text[:6]
    elif text[:4] == 'JSON':
        atype = 'VARCHAR'
    else:
        atype = text
    return [atype, str(is_array)]

def duckdb_decompose(filename:str=None, frame:pd.DataFrame=None, filetype:str=None, path:str="*", limit:int=10000000, recursive:bool=True, root:str="", ignore_errors:bool=True):
    """Decomposes file or data frame structure"""  
    text_ignore = ', ignore_errors=true' if ignore_errors else ''
    if filetype in ['csv', 'tsv']:
        read_func = f"read_csv('{filename}'{text_ignore})"
    elif filetype in ['json', 'jsonl']:
        read_func = f"read_json('{filename}'{text_ignore})"
    else:
        read_func = f"'{filename}'"
    if path == '*':
        if filename is not None:
            data = duckdb.sql(f"summarize select {path} from {read_func} limit {limit}").fetchall()
        else:
            data = duckdb.sql(f"summarize select {path} from frame limit {limit}").fetchall()
    else:
         path_parts = path.split('.')
         query = None
         if len(path_parts) == 1:
            if filename is not None:
                query = f"summarize select unnest(\"{path}\", recursive:=true) from {read_func} limit {limit}"
            else:
                query = f"summarize select unnest(\"{path}\", recursive:=true) from frame limit {limit}"
         elif len(path_parts) == 2:
            if filename is not None:
                query = f"summarize select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from {read_func} limit {limit})"
            else:
                query = f"summarize select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from frame limit {limit})"
         elif len(path_parts) == 3:
            if filename is not None:
                query = f"summarize select unnest(\"{path_parts[2]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from {read_func} limit {limit}))"
            else:
                query = f"summarize select unnest(\"{path_parts[2]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from frame limit {limit}))"                
         elif len(path_parts) == 4:
            if filename is not None:
                query = f"summarize select unnest(\"{path_parts[2]}.{path_parts[3]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from {read_func} limit {limit}))"
            else:
                query = f"summarize select unnest(\"{path_parts[2]}.{path_parts[3]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from frame limit {limit}))"
         data = duckdb.sql(query).fetchall()
    table = []
    for row in data:        
        item = [row[0] if len(root) ==0 else root+ '.' + row[0], ]
        item.extend(column_type_parse(row[1]))
        item.append(str(row[4]))
        item.append(str(row[10]))
        uniq_share = row[4] * 100.0 / row[10] if row[10] > 0 else 0
        item.append('%0.2f' % (uniq_share))
        table.append(item)
        if recursive and item[1] == 'STRUCT':
            subtable = duckdb_decompose(filename, frame, filetype=filetype, path=row[0] if len(root) == 0 else item[0], limit=limit, recursive=recursive, root=item[0], ignore_errors=ignore_errors)
            for subitem in subtable:
                table.append(subitem)
    return table    

def _seek_dict_lists(data, level=0, path=None, candidates=OrderedDict()):
#    print(level, path, candidates)
    for key, value in data.items():
        if isinstance(value, list):
            isobjectlist = False
            for listitem in value[:20]:
                if isinstance(listitem, dict) or isinstance(listitem, OrderedDict):
                    isobjectlist = True
                    break
            if not isobjectlist: 
                continue
            key = path + '.%s' % (key) if path is not None else key
            if key not in candidates.keys():
                candidates[key] = {'key' : key, 'num' : len(value)}
        elif isinstance(value, OrderedDict) or isinstance(value, dict):
            res = _seek_dict_lists(value, level + 1, path + '.' + key if path else key, candidates)
            for k, v in res.items():
                if k not in candidates.keys():
                    candidates[k] = v
        else:
            continue
    return candidates



def _seek_xml_lists(data, level=0, path=None, candidates=OrderedDict()):
    for key, value in data.items():
        if isinstance(value, list):
            key = path + '.%s' % (key) if path is not None else key
            if key not in candidates.keys():
                candidates[key] = {'key' : key, 'num' : len(value)}
        elif isinstance(value, OrderedDict) or isinstance(value, dict):
            res = _seek_xml_lists(value, level + 1, path + '.' + key if path else key, candidates)
            for k, v in res.items():
                if k not in candidates.keys():
                    candidates[k] = v
        else:
            continue
    return candidates



class FieldSchema(BaseModel):
    name: str
    ftype: str
    is_array:bool = False
    description: Optional[str] = None
    sem_type:str = None
    sem_url:str = None


class TableSchema(BaseModel):
    """Table """    
    num_records: int = -1
    num_cols: int = -1
    is_flat:bool = True
    id:Optional[str] = None    
    fields:Optional[list[FieldSchema]] = []
    description:Optional[str] = None

class ReportSchema(BaseModel):
    """Schema of the data file analysis results"""
    filename:str
    file_size:int
    file_type: str
    compression:str = None
    total_tables: int = 1
    total_records:int = -1
    tables:Optional[list[TableSchema]] = []
    metadata:dict = {}
    success:bool = False
    error:str = None


MAX_SAMPLE_SIZE = 200
DELIMITED_FILES = ['csv', 'tsv']
DUCKABLE_FILE_TYPES = ['csv', 'jsonl', 'json', 'parquet']  
DUCKABLE_CODECS  = ['zst', 'gzip', 'raw']


def table_from_objects(objects:list, id:str, objects_limit:int, use_pandas:bool=False, filetype='csv', autodoc:bool=False, lang:str='English'):
    """Reconstructs table schema from list of objects"""
    table = TableSchema(id=id)
    table.num_records = len(objects)
    if autodoc:
       f = io.StringIO()
       writer = csv.writer(f)
       writer.writerows(objects[:MAX_SAMPLE_SIZE])
       table.description = get_description(f.getvalue(), language=lang)
    if use_pandas:
        df = pd.DataFrame(objects)
        columns_raw = duckdb_decompose(frame=df, path='*', limit=objects_limit)
    else:
        tfile = tempfile.NamedTemporaryFile(suffix='.' + filetype, mode='w', encoding='utf8', delete=False)
        tfile.close()
        tfile_real = ZstdFile(tfile.name, mode='w', level_or_option=9)
        wrapper = io.TextIOWrapper(tfile_real, encoding='utf8', write_through=True)
        if filetype == 'csv':
            writer = csv.writer(wrapper) 
            writer.writerows(objects[:objects_limit])
        elif filetype == 'jsonl':
            for row in objects[:objects_limit]:            
                wrapper.write(json.dumps(row) + '\n')
        tfile_real.close()
        # Getting structure
        columns_raw = duckdb_decompose(tfile.name, filetype=filetype, path='*', limit=objects_limit)
        os.remove(tfile.name)
    is_flat = True
    table.num_cols = len(columns_raw)

    for column in columns_raw:                          
        field = FieldSchema(name=column[0], ftype=column[1], is_array=column[2])
        table.fields.append(field)
        if field.ftype == 'STRUCT' or field.is_array: is_flat = False                    
        table.is_flat = is_flat                            
    table.num_records = len(objects)    
    return table




def analyze(filename:str, filetype:str=None, compression:str='raw', objects_limit:int=OBJECTS_ANALYZE_LIMIT, encoding:str=None, scan:bool=True, stats:bool=True, engine:str="auto", use_pandas:bool=False, ignore_errors:bool=True, autodoc:bool=False, lang:str='English'):
    """Analyzes any type of data file and provides miningful insights"""
    fileext = filename.rsplit('.', 1)[-1].lower()
    filesize = os.path.getsize(filename)
    if filetype is None:
        ftype = detect_file_type(filename)
        if ftype['success']:
            filetype = ftype['datatype'].id()            
            if ftype['codec'] is not None:
                compression = ftype['codec'].id()
    # Handling special cases
    if filetype is None and fileext == 'docx': 
        filetype = 'docx'
        
    report = ReportSchema(filename=filename, file_size=filesize, file_type=filetype, compression=compression)    
    if filetype in TEXT_DATA_TYPES:
        if encoding is None: 
            encoding = detect_encoding_any(filename)
            report.metadata['encoding'] = encoding['encoding'] if 'encoding' in encoding.keys() else None
        else:
            report.metadata['encoding'] = encoding
    if scan:        
        if report.file_type in DUCKABLE_FILE_TYPES and report.compression in DUCKABLE_CODECS and engine in ['auto', 'duckdb']:
            # Getting total count
            text_ignore = ', ignore_errors=true' if ignore_errors else ''
            if filetype in ['json', 'jsonl']:
                num_records = duckdb.sql("select count(*) from read_json('%s'%s)" % (filename, text_ignore)).fetchall()[0][0]
            elif filetype in ['csv', 'tsv']:
                num_records = duckdb.sql("select count(*) from read_csv('%s'%s)" % (filename, text_ignore)).fetchall()[0][0]
            else:
                num_records = duckdb.sql("select count(*) from '%s'" % (filename)).fetchall()[0][0]                            
            table = TableSchema(id=os.path.basename(filename))               
            table.num_records = num_records
            report.tables = [table]
            report.total_records = table.num_records
            report.total_tables = 1

            # Getting structure
            columns_raw = duckdb_decompose(filename, filetype=filetype, path='*', limit=objects_limit)
            is_flat = True
            table.num_cols = len(columns_raw)
            for column in columns_raw:                          
                field = FieldSchema(name=column[0], ftype=column[1], is_array=column[2])
                table.fields.append(field)
                if field.ftype == 'STRUCT' or field.is_array: is_flat = False                    
            table.is_flat = is_flat
            sample = duckdb.sql("select * from '%s' limit %d" % (filename, MAX_SAMPLE_SIZE)).fetchall()
            f = io.StringIO()
            writer = csv.writer(f)
            writer.writerows(sample[:MAX_SAMPLE_SIZE])
            table.description = get_description(f.getvalue(), language=lang)
        else:
            if engine == 'duckdb':            
                report.success = False
                report.error = "Not supported file type %s or compression %s" % (report.file_type, report.compression)
            else:
                # Processing MS Word XML files
                if fileext == 'docx':                
                    docx_tables = analyze_docx(filename, extract_data=True)                                   
                    total = 0
                    for dtable in docx_tables:
                        table = table_from_objects(dtable['data'], id=str(dtable['id']), objects_limit=objects_limit, use_pandas=use_pandas, filetype='csv', autodoc=autodoc, lang=lang)
                        total += table.num_records
                        report.tables.append(table)
                    report.total_records = total
                    report.total_tables = len(report.tables)
                elif filetype == 'xlsx':
                    wb = load_workbook(filename)                                   
                    total = 0
                    for sheetname in wb.sheetnames:
                        sheet = wb.get_sheet_by_name(sheetname)
                        objects = []
                        max_num = objects_limit if objects_limit < sheet.max_row else sheet.max_row
                        for n in range(0, max_num):
                            row = next(sheet.iter_rows())
                            tmp = list()
                            for cell in row:
                                tmp.append(str(cell.value))
                            objects.append(tmp)
                        table = table_from_objects(objects, id=sheetname, objects_limit=objects_limit, use_pandas=use_pandas, filetype='csv', autodoc=autodoc, lang=lang)
                        total += table.num_records
                        report.tables.append(table)
                    report.total_records = total
                    report.total_tables = len(report.tables)
                elif filetype == 'xls':
                    wb = xlrd.open_workbook(filename)                                   
                    total = 0
                    for sheetname in wb.sheet_names():
                        sheet = wb.sheet_by_name(sheetname)
                        objects = []
                        max_num = objects_limit if objects_limit < sheet.nrows else sheet.nrows
                        for n in range(0, max_num):
#                            row = next(sheet.get_rows())
                            tmp = list()
                            for i in range(0, sheet.ncols):
                                ct = sheet.cell_type(n, i)
                                cell_value = sheet.cell_value(n, i)
                                get_col = str(cell_value)
                                tmp.append(get_col)    
                            objects.append(tmp)
                        table = table_from_objects(objects, id=sheetname, objects_limit=objects_limit, use_pandas=use_pandas, filetype='csv', autodoc=autodoc, lang=lang)
                        report.tables.append(table)
                        total += table.num_records
                    report.total_records = total
                    report.total_tables = len(report.tables)
                elif filetype == 'xml':
                    fileobj = None
                    codec = None
                    if ftype['codec'] is not None:
                        codec = ftype['codec'](filename, open_it=True)
                        fileobj = codec.fileobj()
                    if fileobj is None:
                        fileobj = open(filename, 'rb')

                    data = xmltodict.parse(fileobj, process_namespaces=False)
                    candidates = _seek_xml_lists(data, level=0)
                    if len(candidates.keys()) == 1:
                        fullkey = str(list(candidates.keys())[0])
                        table = TableSchema(id=fullkey)
                        short_data_key = fullkey.rsplit('.', 1)[-1]
                        objects = get_dict_value(data, keys=fullkey.split('.'))[0]
                        table = table_from_objects(objects, id=fullkey, objects_limit=objects_limit, use_pandas=use_pandas, filetype='jsonl', autodoc=autodoc, lang=lang)
                        report.tables.append(table)
                        report.total_tables = len(report.tables)
                        report.total_records = table.num_records
                    elif len(candidates) > 1:
                        total = 0
                        for fullkey in list(candidates.keys()):
                            table = TableSchema(id=fullkey)
                            short_data_key = fullkey.rsplit('.', 1)[-1]
                            objects = get_dict_value(data, keys=fullkey.split('.'))[0]
                            table = table_from_objects(objects, id=fullkey, objects_limit=objects_limit, use_pandas=use_pandas, filetype='jsonl', autodoc=autodoc, lang=lang)
                            total += table.num_records
                            report.tables.append(table)
                        report.total_records = total
                        report.total_tables = len(report.tables)
                    if codec is not None:
                        codec.close()
                    else:
                        fileobj.close()
                elif filetype == 'json':
                    fileobj = None
                    codec = None
                    if ftype['codec'] is not None:
                        codec = ftype['codec'](filename, open_it=True)
                        fileobj = codec.fileobj()
                    if fileobj is None:
                        fileobj = open(filename, 'rb')

                    data = json.load(fileobj)
                    objects = None
                    if isinstance(data, list):
                        objects = data
                        table = table_from_objects(objects, id=fullkey, objects_limit=objects_limit, use_pandas=use_pandas, filetype='jsonl', autodoc=autodoc, lang=lang)
                        report.tables.append(table)
                        report.total_tables = len(report.tables)
                        report.total_records = table.num_records
                    elif isinstance(data, dict):
                        if len(data.keys()) == 1 and isinstance(data[list(data.keys())[0]], list):
                            objects = data.values()[0]
                            if objects:
                                table = table_from_objects(objects, id=fullkey, objects_limit=objects_limit, use_pandas=use_pandas, filetype='jsonl', autodoc=autodoc, lang=lang)
                                report.tables.append(table)
                                report.total_tables = len(report.tables)
                                report.total_records = table.num_records
                        elif len(data.keys()) > 0:
                            candidates = _seek_dict_lists(data, level=0)
                            if len(candidates) > 0:
                                for fullkey in list(candidates.keys()):
                                    objects = get_dict_value(data, keys=fullkey.split('.'))[0]
                                    table = table_from_objects(objects, id=fullkey, objects_limit=objects_limit, use_pandas=use_pandas, filetype='jsonl', autodoc=autodoc, lang=lang)
                                    total += table.num_records
                                    report.tables.append(table)
                                report.total_records = total
                                report.total_tables = len(report.tables)
                            else:
                                report.append(['JSON type', 'single object'])
                                objects = [data,]
                                table = table_from_objects(objects, id=fullkey, objects_limit=objects_limit, use_pandas=use_pandas, filetype='jsonl', autodoc=autodoc, lang=lang)
                                report.tables.append(table)
                                report.total_tables = len(report.tables)
                                report.total_records = table.num_records
    
    if autodoc and report.total_tables > 0:
        tables = []
        for table in report.tables:            
            fields = []
            for column in table.fields:
                fields.append(column.name)
            descriptions = get_fields_info(fields, language=lang)
            for column in table.fields:
                if column.name in descriptions.keys(): column.description = descriptions[column.name]
    return report





class Analyzer:
    def __init__(self):
        pass


    def analyze(self, filename, options):
        """Analyzes given data file and returns it's parameters"""
        from tabulate import tabulate

        table = None
        encoding = options['encoding'] if 'encoding' in options.keys() else None
        report = analyze(filename, encoding=encoding, engine=options['engine'], use_pandas=options['use_pandas'], autodoc=options['autodoc'], lang=options['lang']) 
        if options['outtype'] == 'json': 
            if options['output'] is not None:
                f = open(options['output'], 'w', encoding='utf8')
                f.write(json.dumps(report.model_dump()))
                f.close()
            else:
                print(json.dumps(report.model_dump(), indent=4, ensure_ascii=False))
        if options['outtype'] == 'yaml': 
            if options['output'] is not None:
                f = open(options['output'], 'w', encoding='utf8')
                f.write(yaml.dump(report.model_dump(), Dumper=yaml.Dumper))
                f.close()
            else:
                print(yaml.dump(report.model_dump(), Dumper=yaml.Dumper))

        elif options['outtype'] == 'markdown':
            raise "Not Implemented"
        else:
            print('Analyzis report')
            headers = ['Attribute', 'Value']
            reptable = []
            reptable.append(['Filename', str(report.filename)])
            reptable.append(['File size', str(report.file_size)])
            reptable.append(['File type', report.file_type])
            reptable.append(['Compression', str(report.compression)])
            reptable.append(['Total tables', str(report.total_tables)])
            reptable.append(['Total records', str(report.total_records)])
            for k, v in report.metadata.items():
                reptable.append([k, v])
            print(tabulate(reptable, headers=headers))

            tabheaders = ['Name', 'Type', 'Is Array', 'Description']
            for rtable in report.tables:
                print()
                table = []
                print("Table %s (items %d, columns %d) structure" % (rtable.id, rtable.num_records, rtable.num_cols))
                for field in rtable.fields:
                    table.append([field.name, field.ftype, str(field.is_array), field.description])#, field.description, field.sem_type, field.sem_url])
                print(tabulate(table, headers=tabheaders))
                print("Summary:")
                print(rtable.description)                
                


# -*- coding: utf8 -*-
from .ruscodes import _check_inn, _check_ogrn
from .commonrules import _validate_email, _validate_url

VALIDATION_RULEMAP = {
    'ru.org.ogrn' : _check_ogrn,
    'ru.org.inn' : _check_inn,
    'common.email' : _validate_email,
    'common.url' : _validate_url
}


# -*- coding: utf8 -*-
import datetime
import bson
import orjson
import csv
import logging
from copy import copy
from .functions import  get_dict_value_deep

OTYPES_MAP = [[type(""), 'string'],
              [type(u""), 'string'],
              [datetime.datetime, 'datetime'],
              [int, 'integer'],
              [bool, 'boolean'],
              [float, 'float'],
              [str, 'string'],
              [bson.int64.Int64, 'integer'],
              [bson.objectid.ObjectId, 'string'],
              [type([]), 'array']
              ]


def merge_schemes(alist, novalue=True):
    """Merges schemes of list of objects and generates final data schema"""
    if len(alist) == 0:
        return None
    obj = alist[0]
    okeys = obj.keys()
    for item in alist[1:]:
        for k in item.keys():
            #            print(obj[k]['type'])
            if k not in okeys:
                obj[k] = item[k]
            elif obj[k]['type'] in ['integer', 'float', 'string', 'datetime']:
                if not novalue:
                    obj[k]['value'] += item[k]['value']
            elif obj[k]['type'] == 'dict':
                if not novalue:
                    obj[k]['value'] += item[k]['value']
                if 'schema' in item[k].keys():
                    obj[k]['schema'] = merge_schemes([obj[k]['schema'], item[k]['schema']])
            elif obj[k]['type'] == 'array':
                #                if 'subtype' not in obj[k].keys():
                #                   logging.info(str(obj[k]))
                if 'subtype' in obj[k].keys() and obj[k]['subtype'] == 'dict':
                    if not novalue:
                        obj[k]['value'] += item[k]['value']
                    if 'schema' in item[k].keys():
                        obj[k]['schema'] = merge_schemes([obj[k]['schema'], item[k]['schema']])
                else:
                    if not novalue:
                        obj[k]['value'] += item['value']
    return obj


def get_schemes(alist):
    """Generates schemas for each object"""
    results = []
    for o in alist:
        results.append(get_schema(o))
    return results

def get_schema(obj, novalue=True):
    """Generates schema from object"""
    result = {}
    for k in obj.keys():
        tt = type(obj[k])
        if obj[k] is None:
            result[k] = {'type': 'string', 'value' : 1}
        elif tt == type("") or tt == type(u"") or isinstance(obj[k], str):
            result[k] = {'type': 'string', 'value' : 1}
        elif isinstance(obj[k], str):
            result[k] = {'type': 'string', 'value': 1}
        elif tt == datetime.datetime:
            result[k] = {'type': 'datetime', 'value' : 1}
        elif tt == bool:
            result[k] = {'type': 'boolean', 'value' : 1}
        elif tt == float:
            result[k] = {'type': 'float', 'value' : 1}
        elif tt == int:
            result[k] = {'type': 'integer', 'value' : 1}
        elif tt == bson.int64.Int64:
            result[k] = {'type': 'integer', 'value' : 1}
        elif tt == bson.objectid.ObjectId:
            result[k] = {'type': 'string', 'value' : 1}
        elif tt == type({}):
            result[k] = {'type': 'dict', 'value' : 1, 'schema' : get_schema(obj[k])}
        elif tt == type([]):
            result[k] = {'type': 'array', 'value' : 1}
            if len(obj[k]) == 0:
                result[k]['subtype'] = 'string'
            else:
                found = False
                for otype, oname in OTYPES_MAP:
                    if type(obj[k][0]) == otype:
                        result[k]['subtype'] = oname
                        found = True
                if not found:
                    if type(obj[k][0]) == type({}):
                        result[k]['subtype'] = 'dict'
                        result[k]['schema'] =  merge_schemes(get_schemes(obj[k]))
                    else:
                        logging.info("Unknown object %s type %s" % (k, str(type(obj[k][0]))))
        else:
            logging.info("Unknown object %s type %s" % (k, str(type(obj[k]))))
            result[k] = {'type': 'string', 'value' : 1}
        if novalue:
            del result[k]['value']
    return result

def extract_keys(obj, parent=None, text=None, level=1):
    """Extracts keys"""
    text = ''
    if not parent:
        text = "'schema': {\n"
    for k in obj.keys():
        if type(obj[k]) == type({}):
            text += "\t" * level + "'%s' : {'type' : 'dict', 'schema' : {\n" % (k)
            text += extract_keys(obj[k], k, text, level+1)
            text += "\t" * level + "}},\n"
        elif type(obj[k]) == type([]):
            text += "\t" * level + "'%s' : {'type' : 'list', 'schema' : { 'type' : 'dict', 'schema' : {\n" % (k)
            if len(obj[k]) > 0:
                item = obj[k][0]
                if type(item) == type({}):
                    text += extract_keys(item, k, text, level+1)
                else:
                    text += "\t" * level + "'%s' : {'type' : 'string'},\n" % (k)
            text += "\t" * level + "}}},\n"
        else:
            logging.info(str(type(obj[k])))
            text += "\t" * level + "'%s' : {'type' : 'string'},\n" % (k)
    if not parent:
        text += "}"
    return text

def __get_filetype_by_ext(filename):
    ext = filename.rsplit('.', 1)[-1].lower()
    if ext in ['bson', 'json', 'csv', 'jsonl']:
        return ext
    return filename


def generate_scheme_from_file(filename=None, fileobj=None, filetype='bson', alimit=1000, verbose=0, encoding='utf8', delimiter=",", quotechar='"'):
    """Generates schema of the data BSON file"""
    if not filetype and filename is not None:
        filetype = __get_filetype_by_ext(filename)
    datacache = []
    if filetype == 'bson':
        if filename:
            source = open(filename, 'rb')
        else:
            source = fileobj
        n = 0
        for r in bson.decode_file_iter(source):
            n += 1
            if n > alimit:
                break
            datacache.append(r)
        if filename:
            source.close()
    elif filetype == 'jsonl':
        if filename:
            source = open(filename, 'r', encoding=encoding)
        else:
            source = fileobj
        n = 0
        for r in source:
            n += 1
            if n > alimit:
                break
            datacache.append(orjson.loads(r))
        if filename:
            source.close()
    elif filetype == 'csv':
        if filename:
            source = open(filename, 'r', encoding=encoding)
        else:
            source = fileobj
        n = 0
        reader = csv.DictReader(source, quotechar=quotechar, delimiter=delimiter, quoting=csv.QUOTE_ALL)
        for r in reader:
            n += 1
            if n > alimit:
                break
            datacache.append(r)
        if filename:
            source.close()
    n = 0
    scheme = None
    for r in datacache:
        n += 1
        if scheme is None:
            scheme = get_schema(r)
        else:
            scheme = merge_schemes([scheme, get_schema(r)])
    return scheme

def schema2fieldslist(schema, prefix=None, predefined=None, sample=None):
    """Converts data schema to the fields list"""
    fieldslist = []
    for k in schema.keys():
        if prefix is None:
            name = k
        else:
            name = '.'.join(['.'.join(prefix.split('.')), k])
        try:
            sampledata = get_dict_value_deep(sample, name) if sample else ''
        except:
            sampledata = ''
        if 'schema' not in schema[k].keys():
            if schema[k]['type'] != 'array':
                field = {'name' : name, 'type': schema[k]['type'], 'description' : '', 'sample' : sampledata, 'class' : ""}
            else:
                field = {'name': name, 'type': 'list of [%s]' % schema[k]['type'], 'description' : '', 'sample' : sampledata, 'class' : ""}
            if predefined:
                if name in predefined.keys():
                    field['description'] = predefined[name]['text']
                    if predefined[name]['class']:
                        field['class'] = predefined[name]['class']
                elif k in predefined.keys():
                    field['description'] = predefined[k]['text']
                    if predefined[k]['class']:
                        field['class'] = predefined[k]['class']
            if field['type'] == 'datetime':
                field['class'] = 'datetime'
            fieldslist.append(field)
        else:
            if prefix is not None:
                subprefix = copy(prefix) + '.' + k
#                subprefix.append(k)
            else:
                subprefix = k
            if schema[k]['type'] == 'dict':
                field =  {'name' : name, 'type': schema[k]['type'], 'description' : '', 'sample' : '', 'class' : ''}
                if predefined:
                    if name in predefined.keys():
                        field['description'] = predefined[name]['text']
                        if predefined[name]['class']:
                            field['class'] = predefined[name]['class']
                    elif k in predefined.keys():
                        field['description'] = predefined[k]['text']
                        if predefined[k]['class']:
                            field['class'] = predefined[k]['class']
                fieldslist.append(field)
                fieldslist.extend(schema2fieldslist(schema[k]['schema'], prefix=subprefix, predefined=predefined, sample=sample))
            elif schema[k]['type'] == 'array':
                field = {'name': name, 'type': 'list of [%s]' % schema[k]['type'], 'description' : '', 'sample' : '', 'class' : ''}
                if predefined:
                    if name in predefined.keys():
                        field['description'] = predefined[name]['text']
                        if predefined[name]['class']:
                            field['class'] = predefined[name]['class']
                    elif k in predefined.keys():
                        field['description'] = predefined[k]['text']
                        if predefined[k]['class']:
                            field['class'] = predefined[k]['class']
                fieldslist.append(field)
                fieldslist.extend(schema2fieldslist(schema[k]['schema'], prefix=subprefix, sample=sample))
    return fieldslist



# -*- coding: utf-8 -*-
from ..constants import BINARY_FILE_TYPES
from ..utils import get_file_type, get_option, detect_encoding, detect_delimiter
import csv
import jsonlines
import bson
import logging
import io

SUPPORTED_COMPRESSION = {'gz': True, 'zip': True, 'xz': False, '7z': False,  'lz4': False, 'bz2' : True}
import gzip
from zipfile import ZipFile
from lzma import LZMAFile
from bz2 import BZ2File

try:
    import lz4
    SUPPORTED_COMPRESSION['lz4'] = True
except ImportError:
    pass

try:
    import py7zr
    SUPPORTED_COMPRESSION['7z'] = True
except ImportError:
    pass





DEFAULT_ENCODING = 'utf8'
DEFAULT_DELIMITER = ','

class IterableData:
    """Iterable data reader (CSV/JSON lines, BSON"""
    def __init__(self, filename, options={}, autodetect=True, autodetect_limit=100000):
        """Creates iterable object from CSV, JSON lines and other iterable files.
        """
        self.autodetect = autodetect
        self.autodetect_limit = autodetect_limit
        self.options = options
        self.archiveobj = None
        self.fileobj = None
        self.binary = False
        self.delimiter = get_option(options, 'delimiter')
        self.init(filename, options)
        pass


    def init(self, filename, options):
        f_type = get_file_type(filename) if options['format_in'] is None else options['format_in']
        self.encoding = get_option(options, 'encoding')
        self.filetype = f_type
        ext = filename.rsplit('.', 1)[-1].lower()
        self.ext = ext
        if ext in SUPPORTED_COMPRESSION.keys():
            self.binary = True
            self.mode = 'rb' if self.filetype in BINARY_FILE_TYPES else 'r'
            if ext == 'gz':
                self.fileobj = gzip.open(filename, self.mode)
            elif ext == 'bz2':
                self.fileobj = BZ2File(filename, self.mode)
            elif ext == 'xz':
                self.fileobj = LZMAFile(filename, self.mode)
            elif ext == 'zip':
                self.archiveobj = ZipFile(filename, mode='r')
                fnames = self.archiveobj.namelist()
                self.fileobj = self.archiveobj.open(fnames[0], self.mode)
            else:
                raise NotImplementedError
        else:
            if f_type in BINARY_FILE_TYPES:
                self.fileobj = open(filename, 'rb')
            else:
                if 'encoding' in options.keys() and options['encoding']:
                    encoding = get_option(options, 'encoding')
                else:
                    if self.autodetect:
                        detected_enc = detect_encoding(filename, limit=self.autodetect_limit)
                        encoding = detected_enc['encoding'] if detected_enc else DEFAULT_ENCODING
                        if f_type == 'csv':
                            detected_del = detect_delimiter(filename, encoding)
                            delimiter = detected_del if detected_del else DEFAULT_DELIMITER
                            self.delimiter = delimiter
                    else:
                        encoding = DEFAULT_ENCODING
                        if f_type == 'csv':
                            delimiter = DEFAULT_DELIMITER
                            self.delimiter = delimiter
                    logging.debug('Detected encoding %s' % (detected_enc['encoding']))
                self.encoding = encoding
                self.fileobj = open(filename, 'r', encoding=encoding)

    def init_orig(self, filename, options):
        f_type = get_file_type(filename) if options['format_in'] is None else options['format_in']
        encoding = get_option(options, 'encoding')
        self.filetype = f_type
        if options['zipfile']:
            z = ZipFile(filename, mode='r')
            fnames = z.namelist()
            if f_type in BINARY_FILE_TYPES:
                self.fileobj = z.open(fnames[0], 'rb')
            else:
                self.fileobj = z.open(fnames[0], 'r')
        else:
            if f_type in BINARY_FILE_TYPES:
                self.fileobj = open(filename, 'rb')
            else:
                if 'encoding' in options.keys() and options['encoding']:
                    encoding = get_option(options, 'encoding')
                else:
                    if self.autodetect:
                        detected_enc = detect_encoding(filename, limit=self.autodetect_limit)
                        encoding = detected_enc['encoding'] if detected_enc else DEFAULT_ENCODING
                        if f_type == 'csv':
                            detected_del = detect_delimiter(filename, encoding)
                            delimiter = detected_del if detected_del else DEFAULT_DELIMITER
                            self.delimiter = delimiter
                    else:
                        encoding = DEFAULT_ENCODING
                        if f_type == 'csv':
                            delimiter = DEFAULT_DELIMITER
                    logging.debug('Detected encoding %s' % (detected_enc['encoding']))
                self.encoding = encoding
                self.fileobj = open(filename, 'r', encoding=encoding)

    def iter(self):
        if self.filetype == 'csv':
            if self.binary:
                obj = io.TextIOWrapper(self.fileobj, encoding=self.encoding)
                reader = csv.DictReader(obj, delimiter=self.delimiter)
            else:
                reader = csv.DictReader(self.fileobj, delimiter=self.delimiter)
            return iter(reader)
        elif self.filetype == 'jsonl':
            return jsonlines.Reader(self.fileobj)
        elif self.filetype == 'bson':
            return bson.decode_file_iter(self.fileobj)

    def close(self):
        """Closes file object and archive file object if it exists"""
        if self.fileobj is not None:
            self.fileobj.close()
        if self.archiveobj is not None:
            self.archiveobj.close()



class BSONWriter:
    def __init__(self, fileobj):
        self.fo = fileobj

    def write(self, item):
        rec = bson.BSON.encode(item)
        self.fo.write(rec)


class DataWriter:
    """Data writer (CSV/JSON lines, BSON"""
    def __init__(self, fileobj, filetype, delimiter=',', fieldnames=None):
        """Creates iterable object from CSV, JSON lines or BSON file.
        """
        self.filetype = filetype
        self.fieldnames = fieldnames
        self.fileobj = fileobj
        if self.filetype == 'csv':
            self.writer = csv.DictWriter(self.fileobj, delimiter=delimiter, fieldnames=fieldnames)
        elif self.filetype == 'jsonl':
            self.writer = jsonlines.Writer(self.fileobj)
        elif self.filetype == 'bson':
            self.writer = BSONWriter(self.fileobj)
        pass


#    def write_item(self, row):

    def write_items(self, outdata):
        if len(outdata) == 0:
            return
        if self.filetype == 'csv':
            self.writer.writeheader()
            if isinstance(outdata[0], str):
                for rawitem in outdata:
                    item = {self.fieldnames[0]: rawitem}
                    self.writer.writerow(item)
            elif type(outdata[0]) == type([]):
                for rawitem in outdata:
                    item = dict(zip(self.fieldnames, rawitem))
                    self.writer.writerow(item)
            else:
                self.writer.writerows(outdata)
        elif self.filetype in ['jsonl', 'bson']:
            # If our data is just array of strings, we just transform it to dict
            if isinstance(outdata[0], str):
                for rawitem in outdata:
                    item = {self.fieldnames[0]: rawitem}
                    self.writer.write(item)
#                    handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
            elif type(outdata[0]) == type([]):
                for rawitem in outdata:
                    item = dict(zip(self.fieldnames, rawitem))
                    self.writer.write(item)
#                    handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
            else:
                for item in outdata:
                    self.writer.write(item)
#                    handle.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))


if __name__ == "__main__":
    f = open('outtest.jsonl', 'w')
    writer = DataWriter(f, filetype='jsonl', fieldnames=['name', 'value'])
    writer.write_items([{'name' : 'Cat', 'value' : 15}])
    pass



# -*- coding: utf8 -*-
import sys
import orjson
import logging
#from xmlr import xmliter
from ..utils import get_option, dict_generator
from runpy import run_path
from iterable.helpers.detect import open_iterable

ITERABLE_OPTIONS_KEYS = ['tagname', 'delimiter', 'encoding', 'start_line', 'page']


def get_iterable_options(options):
    out = {}
    for k in ITERABLE_OPTIONS_KEYS:
        if k in options.keys():
            out[k] = options[k]
    return out            

DEFAULT_HEADERS_DETECT_LIMIT = 1000



class Transformer:
    def __init__(self):
        pass


    def script(self, fromfile, options={}):
        """Run certain script against selected file"""

        script = run_path(options['script'])
        __process_func = script['process']

        iterableargs = get_iterable_options(options)
        read_iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)

        limit = DEFAULT_HEADERS_DETECT_LIMIT

        keys = []
        n = 0
        for item in read_iterable:
            if limit and n > limit:
                break
            item = __process_func(item)
            n += 1
            dk = dict_generator(item)
            for i in dk:
                k = ".".join(i[:-1])
                if k not in keys:
                    keys.append(k)

        read_iterable.close()
        read_iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)


        write_to_iterable = False
        to_file = get_option(options, 'output')
        if to_file:            
            write_to_iterable = True
            write_iterable = open_iterable(to_file, mode='w', iterableargs={'keys' : keys}) 
        n = 0
        for r in read_iterable:
            n += 1
            if n % 10000 == 0:                                                                                                                 
                logging.info('apply script: processing %d records of %s' % (n, fromfile))
            item = __process_func(r)
            if write_to_iterable:
                write_iterable.write(item)
            else:
                sys.stdout.write(orjson.dumps(item, option=orjson.OPT_APPEND_NEWLINE).decode('utf8'))
       
        logging.debug('select: %d records processed' % (n))
        read_iterable.close()
        if write_to_iterable:
            write_iterable.close()





#!/usr/bin/env python
# -*- coding: utf8 -*-
import typer
import logging

from .cmds.converter import Converter
from .cmds.selector import Selector
from .cmds.transformer import Transformer
from .cmds.analyzer import Analyzer
from .cmds.statistics import StatProcessor
from .cmds.textproc import TextProcessor
from .cmds.validator import Validator
from .cmds.schemer import Schemer
from .cmds.query import DataQuery
from .cmds.ingester import Ingester

import glob

DEFAULT_BATCH_SIZE = 1000

app = typer.Typer()

#logging.getLogger().addHandler(logging.StreamHandler())
logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    level=logging.INFO)

def enableVerbose():
    logging.basicConfig(
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        level=logging.INFO)

@app.command()
def convert(input:str, output:str, delimiter:str=',', compression:str='brotli', encoding:str='utf8', verbose:bool=False, flatten:bool=False, prefix_strip:bool=True, fields:str=None, start_line:int=0, skip_end_rows:int=0, start_page:int=0, tagname:str=None, format_in:str=None, format_out:str=None, zipfile:bool=False):
    """Converts one file to another. Supports XML, CSV, JSON, BSON"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['compression'] = compression
    options['flatten'] = flatten
    options['encoding'] = encoding
    options['prefix_strip'] = prefix_strip
    options['start_line'] = start_line
    options['skip_end_rows'] = skip_end_rows
    options['start_page'] = start_page
    options['tagname'] = tagname
    options['fields'] = fields
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    acmd = Converter()
    acmd.convert(input, output, options)
    pass

@app.command()
def convertold(input:str, output:str, delimiter:str=',', compression:str='brotli', encoding:str='utf8', verbose:bool=False, flatten:bool=False, prefix_strip:bool=True, fields:str=None, start_line:int=0, skip_end_rows:int=0, start_page:int=0, tagname:str=None, format_in:str=None, format_out:str=None, zipfile:bool=False):
    """Converts one file to another. Supports XML, CSV, JSON, BSON (old implementation)"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['compression'] = compression
    options['flatten'] = flatten
    options['encoding'] = encoding
    options['prefix_strip'] = prefix_strip
    options['start_line'] = start_line
    options['skip_end_rows'] = skip_end_rows
    options['start_page'] = start_page
    options['tagname'] = tagname
    options['fields'] = fields
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    acmd = Converter()
    acmd.convert_old(input, output, options)
    pass

@app.command()
def uniq(input:str, output:str=None, fields:str=None, delimiter:str=',', encoding:str=None, verbose:bool=False, format_in:str=None, format_out:str=None, zipfile:bool=False, filter:str=None):
    """Returns all unique files of certain field(s)"""
    if verbose:
        enableVerbose()
    options = {}
    options['output'] = output
    options['fields'] = fields
    options['delimiter'] = delimiter
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['filter'] = filter
    acmd = Selector()
    acmd.uniq(input, options)
    pass


@app.command()
def headers(input:str, output:str=None, fields:str=None, delimiter:str=',', encoding:str=None, limit:int=10000, verbose:bool=False, format_in:str=None, format_out:str=None, zipfile:bool=False, filter:str=None):
    """Returns fieldnames of the file. Supports XML, CSV, JSON, BSON"""
    if verbose:
        enableVerbose()
    options = {}
    options['output'] = output
    options['delimiter'] = delimiter
    options['encoding'] = encoding
    options['limit'] = limit
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    acmd = Selector()
    acmd.headers(input, options)
    pass

@app.command()
def stats(input:str, output:str=None, dictshare:int=None, format_in:str=None, format_out:str=None, delimiter:str=None, verbose:bool=False, zipfile:bool=False, checkdates:bool=True, encoding:str=None):
    """Returns detailed stats on selected dataset"""
    if verbose:
        enableVerbose()
    options = {}
    options['output'] = output
    options['dictshare'] = dictshare
    options['zipfile'] = zipfile
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['delimiter'] = delimiter
    options['checkdates'] = checkdates
    options['encoding'] = encoding
    options['verbose'] = verbose
    acmd = StatProcessor(nodates=not checkdates)
    acmd.stats(input, options)
    pass

def flatten(input:str, output:str=None, delimiter:str=',', encoding:str='utf8', format_in:str=None, filter:str=None, verbose:bool=False):
    """Flatten data records. Write them as one value per row"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['filter'] = filter
    acmd = TextProcessor()
    acmd.flatten(input, options)
    pass


@app.command()
def frequency(input:str, output:str=None, fields:str=None, delimiter:str=",", encoding:str=None, verbose:bool=False, format_in:str=None, format_out:str=None, zipfile:bool=False, filter:str=None):
    """Field value frequency calc"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['filter'] = filter
    acmd = Selector()
    acmd.frequency(input, options)
    pass

@app.command()
def select(input:str, output:str=None, fields:str=None, delimiter:str=",", encoding:str=None, verbose:bool=False, format_in:str=None, format_out:str=None, zipfile:bool=False, filter:str=None):
    """Select or re-order columns from file. Supports CSV, JSONl, BSON"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['filter'] = filter
    acmd = Selector()
    acmd.select(input, options)
    pass


@app.command()
def split(input:str, output:str=None, fields:str=None, delimiter:str=',', encoding:str="utf8", verbose:bool=False, format_in:str=None, zipfile:bool=False, gzipfile:str=None, chunksize:int=10000, filter:str=None, dirname:str=None):
    """Splits the given file with data into chunks."""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['gzipfile'] = gzipfile
    options['chunksize'] = chunksize
    options['filter'] = filter
    options['dirname'] = dirname
    acmd = Selector()
    acmd.split(input, options)
    pass

@app.command()
def validate(input:str, output:str=None, fields:str=None, delimiter:str=',', encoding:str='utf8', verbose:bool=False, format_in:str=None, zipfile:bool=False, rule:str=None, filter:str=None, mode:str="invalid"):
    """Validates selected field against validation rule"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['filter'] = filter
    options['rule'] = rule
    options['mode'] = mode
    acmd = Validator()
    acmd.validate(input, options)
    pass

@app.command()
def apply(input:str, output:str=None, fields:str=None, delimiter:str=",", encoding:str='utf8', verbose:bool=False, format_in:str=None, zipfile:bool=False, script:str=None, filter:str=None):
    """Runs script against each record of input file"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['filter'] = filter
    options['script'] = script
    acmd = Transformer()
    acmd.script(input, options)
    pass


@app.command()
def scheme(input:str, output:str=None, delimiter:str=',', encoding:str='utf8', verbose:bool=False, format_in:str=None, zipfile:bool=False, stype:str='cerberus'):
    """Generate data schema from file"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['zipfile'] = zipfile
    options['stype'] = stype
    acmd = Schemer()
    acmd.generate_scheme(input, options)
    pass



@app.command()
def analyze(input:str, verbose:bool=False, engine:str="auto", use_pandas:bool=False, outtype:str="text", output:str=None, autodoc:bool=False, lang:str="English"):
    """Analyzes given data file and returns human readable insights about it"""
    if verbose:
        enableVerbose()
    options = {}
    options['engine'] = engine
    options['use_pandas'] = use_pandas
    options['outtype'] = outtype
    options['output'] = output
    options['autodoc'] = autodoc
    options['lang'] = lang
    acmd = Analyzer()
    acmd.analyze(input, options)
    pass

@app.command()
def schema(input:str, verbose:bool=False, outtype:str="text", output:str=None, autodoc:bool=False, lang:str="English"):
    """Schema extraction"""
    if verbose:
        enableVerbose()
    options = {}
    options['outtype'] = outtype
    options['output'] = output
    options['autodoc'] = autodoc
    options['lang'] = lang
    acmd = Schemer()
    acmd.extract_schema(input, options)
    pass

@app.command()
def schema_bulk(input:str, verbose:bool=False, outtype:str="text", output:str=None, mode:str="distinct", autodoc:bool=False, lang:str="English"):
    """Schema extraction from many files. Default mode is 'distinct' that creates unique schema files per schema, alternative is 'perfile' that creates a schema per file with same names"""
    if verbose:
        enableVerbose()
    options = {}
    options['outtype'] = outtype
    options['output'] = output
    options['mode'] = mode
    options['autodoc'] = autodoc
    options['lang'] = lang
    acmd = Schemer()
    acmd.extract_schema_bulk(input, options)
    pass

@app.command()
def ingest(input:str, uri:str, db:str, table:str, verbose:bool=False, batch:int=DEFAULT_BATCH_SIZE, dbtype:str="mongodb", totals:bool=False, drop:bool=False, timeout:int=-30, skip:int=None, api_key:str=None, doc_id:str=None):
    """Data ingester"""
    if verbose:
        enableVerbose()
    options = {}
    options['dbtype'] = dbtype
    options['skip'] = skip
    options['drop'] = drop
    options['totals'] = totals
    options['doc_id'] =  doc_id
    options['api_key'] =  api_key
    options['timeout'] =  timeout
    acmd = Ingester(batch)
    files = glob.glob(input.strip("'"))    
    acmd.ingest(files, uri, db, table, options)
    pass



@app.command()
def query(input:str, output:str=None, fields:str=None, delimiter:str=',', encoding:str=None, verbose:bool=False, format_in:str=None, format_out:str=None, zipfile:bool=False, query:str=None):
    """Query data using mistql (experimental, require mistql). Use 'pip install mistql' to install"""
    if verbose:
        enableVerbose()
    options = {}
    options['delimiter'] = delimiter
    options['fields'] = fields
    options['output'] = output
    options['encoding'] = encoding
    options['format_in'] = format_in
    options['format_out'] = format_out
    options['zipfile'] = zipfile
    options['query'] = query
    acmd = DataQuery()
    acmd.query(input, options)
    pass




if __name__ == '__main__':
    app()




import pandas as pd
from undatum.cmds.analyzer import duckdb_decompose

DATA = [
    {"foo": 1, "bar": "some string", "baz": 1.23},
    {"foo": 2, "bar": "some other string", "baz": 2.34},
    {"foo": 3, "bar": "yet another string", "baz": 3.45},
]

df = pd.DataFrame(DATA)

print(duckdb_decompose(frame=df))


# -*- coding: utf8 -*-
from ..utils import get_file_type, get_option
from iterable.helpers.detect import open_iterable

#STAT_READY_DATA_FORMATS = ['jsonl', 'bson', 'csv']

ITERABLE_OPTIONS_KEYS = ['tagname', 'delimiter', 'encoding', 'start_line', 'page']


def get_iterable_options(options):
    out = {}
    for k in ITERABLE_OPTIONS_KEYS:
        if k in options.keys():
            out[k] = options[k]
    return out            


def get_keys(adict, prefix=None):
    keys = {}
    for k, v in adict.items():
        fullk = '.'.join([prefix, k]) if prefix else k
        keys[fullk] = 1
        if type(v) == type({}):
            for ak in get_keys(v, fullk):
                keys[ak] = 1
        elif type(v) == type([]):
            for item in v:
                if type(item) == type({}):
                    for ak in get_keys(item, fullk):
                        keys[ak] = 1
        else:
            print((u'%s\t%s' % (fullk, str(v))))
    return keys


class TextProcessor:
    def __init__(self):
        pass

    def flatten(self, filename, options):
        """Flatten the data. One field - one line"""
        get_file_type(filename) if options['format_in'] is None else options['format_in']
        iterableargs = get_iterable_options(options)
        iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)
        get_option(options, 'output')
        i = 0
        for rec in iterable:
            allkeys = {}
            i += 1
            for k in get_keys(rec):
                v = allkeys.get(k, 0)
                allkeys[k] = v + 1
            for k, v in allkeys.items():
                print('\t'.join([k, str(v)]))




# -*- coding: utf8 -*-
from ..utils import get_option, dict_generator, guess_datatype
from ..constants import DEFAULT_DICT_SHARE
import logging
from qddate import DateParser
#from ..common.iterable import IterableData
from iterable.helpers.detect import open_iterable

#STAT_READY_DATA_FORMATS = ['jsonl', 'bson', 'csv']

ITERABLE_OPTIONS_KEYS = ['tagname', 'delimiter', 'encoding', 'start_line', 'page']


def get_iterable_options(options):
    out = {}
    for k in ITERABLE_OPTIONS_KEYS:
        if k in options.keys():
            out[k] = options[k]
    return out            


class StatProcessor:
    def __init__(self, nodates=True):
        if nodates:
            self.qd = None
        else:
            self.qd = DateParser(generate=True)
        pass

    def stats(self, fromfile, options):
        """Produces statistics and structure analysis of JSONlines, BSON or CSV file and produces stats"""
        from rich import print
        from rich.table import Table

        iterableargs = get_iterable_options(options)
        iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)
        dictshare = get_option(options, 'dictshare')

        if dictshare and dictshare.isdigit():
            dictshare = int(dictshare)
        else:
            dictshare = DEFAULT_DICT_SHARE

        profile = {'version': 1.0}
        fielddata = {}
        fieldtypes = {}

        #    data = json.load(open(profile['filename']))
        count = 0
        nfields = 0

        # process data items one by one
        logging.debug('Start processing %s' % (fromfile))
        for item in iterable:
            count += 1
            dk = dict_generator(item)
            if count % 1000 == 0: logging.debug('Processing %d records of %s' % (count, fromfile))
            for i in dk:
                #            print(i)
                k = '.'.join(i[:-1])
                if len(i) == 0: continue
                if i[0].isdigit(): continue
                if len(i[0]) == 1: continue
                v = i[-1]
                if k not in list(fielddata.keys()):
                    fielddata[k] = {'key': k, 'uniq': {}, 'n_uniq': 0, 'total': 0, 'share_uniq': 0.0,
                                    'minlen': None, 'maxlen': 0, 'avglen': 0, 'totallen': 0}
                fd = fielddata[k]
                uniqval = fd['uniq'].get(v, 0)
                fd['uniq'][v] = uniqval + 1
                fd['total'] += 1
                if uniqval == 0:
                    fd['n_uniq'] += 1
                    fd['share_uniq'] = (fd['n_uniq'] * 100.0) / fd['total']
                fl = len(str(v))
                if fd['minlen'] is None:
                    fd['minlen'] = fl
                else:
                    fd['minlen'] = fl if fl < fd['minlen'] else fd['minlen']
                fd['maxlen'] = fl if fl > fd['maxlen'] else fd['maxlen']
                fd['totallen'] += fl
                fielddata[k] = fd
                if k not in list(fieldtypes.keys()):
                    fieldtypes[k] = {'key': k, 'types': {}}
                fd = fieldtypes[k]
                thetype = guess_datatype(v, self.qd)['base']
                uniqval = fd['types'].get(thetype, 0)
                fd['types'][thetype] = uniqval + 1
                fieldtypes[k] = fd
        #        print count
        for k, v in list(fielddata.items()):
            fielddata[k]['share_uniq'] = (v['n_uniq'] * 100.0) / v['total']
            fielddata[k]['avglen'] = v['totallen'] / v['total']
        profile['count'] = count
        profile['num_fields'] = nfields
        dictkeys = []
        dicts = {}
        #        print(profile)
        profile['fields'] = []
        for fd in list(fielddata.values()):
            #            print(fd['key'])  # , fd['n_uniq'], fd['share_uniq'], fieldtypes[fd['key']]
            field = {'key': fd['key'], 'is_uniq': 0 if fd['share_uniq'] < 100 else 1}
            profile['fields'].append(field)
            if fd['share_uniq'] < dictshare:
                dictkeys.append(fd['key'])
                dicts[fd['key']] = {'items': fd['uniq'], 'count': fd['n_uniq'],
                                    'type': 'str'}  # TODO: Shouldn't be "str" by default
        #            for k, v in fd['uniq'].items():
        #                print fd['key'], k, v
        profile['dictkeys'] = dictkeys

        finfields = {}
        for k, v in list(fielddata.items()):
            del v['uniq']
            fielddata[k] = v
        profile['debug'] = {'fieldtypes': fieldtypes.copy(), 'fielddata': fielddata}
        for fd in list(fieldtypes.values()):
            fdt = list(fd['types'].keys())
            if 'empty' in fdt:
                del fd['types']['empty']
            if len(list(fd['types'].keys())) != 1:
                ftype = 'str'
            else:
                ftype = list(fd['types'].keys())[0]
            finfields[fd['key']] = ftype

        profile['fieldtypes'] = finfields
        table = []
        for fd in list(fielddata.values()):
            field = [fd['key'], ]
            field.append(finfields[fd['key']])
            field.append(True if fd['key'] in dictkeys else False)
            field.append(False if fd['share_uniq'] < 100 else True)
            field.append(fd['n_uniq'])
            field.append(fd['share_uniq'])
            field.append(fd['minlen'])
            field.append(fd['maxlen'])
            field.append(fd['avglen'])
            table.append(field)
        headers = ('key', 'ftype', 'is_dictkey', 'is_uniq', 'n_uniq', 'share_uniq', 'minlen', 'maxlen', 'avglen')
        reptable = Table(title="Statistics")
        reptable.add_column(headers[0], justify="left", style="magenta")
        for key in headers[1:-1]:
            reptable.add_column(key, justify="left", style="cyan", no_wrap=True)
        reptable.add_column(headers[-1], justify="right", style="cyan")
        for row in table:
            reptable.add_row(*map(str, row))
        print(reptable)




# -*- coding: utf8 -*-
from email.utils import parseaddr
import validators.url

def _validate_email(s):
    return '@' in parseaddr(s)[1]

def _validate_url(s):
    r = validators.url(s)
    return (r is True)


# -*- coding: utf8 -*-
import csv
import json
import orjson
import logging
import pandas
# from xmlr import xmliter
import xml.etree.ElementTree as etree
from collections import defaultdict

import bson
from bson import ObjectId
from xlrd import open_workbook as load_xls
from ..utils import get_file_type, get_option, dict_generator
from iterable.helpers.detect import open_iterable
from tqdm import tqdm

ITERABLE_OPTIONS_KEYS = ['tagname', 'delimiter', 'encoding', 'start_line', 'page']

DEFAULT_BATCH_SIZE = 50000

def get_iterable_options(options):
    out = {}
    for k in ITERABLE_OPTIONS_KEYS:
        if k in options.keys():
            out[k] = options[k]
    return out            



PREFIX_STRIP = True
PREFIX = ""

LINEEND = '\n'.encode('utf8')

def df_to_pyorc_schema(df):
    """Extracts column information from pandas dataframe and generate pyorc schema"""
    struct_schema = []
    for k, v in df.dtypes.to_dict().items():
        v = str(v)
        if v == 'float64':
            struct_schema.append('%s:float' % (k))
        elif v == 'float32':
            struct_schema.append('%s:float' % (k))
        elif v == 'datetime64[ns]':
            struct_schema.append('%s:timestamp' % (k))
        elif v == 'int32':
            struct_schema.append('%s:int' % (k))
        elif v == 'int64':
            struct_schema.append('%s:int' % (k))
        else:
            struct_schema.append('%s:string' %(k))
    return struct_schema


def __copy_options(user_options, default_options):
    """If user provided option so we use it, if not, default option value should be used"""
    for k in default_options.keys():
        if k not in user_options.keys():
            user_options[k] = default_options[k]
    return user_options


def etree_to_dict(t, prefix_strip=True):
    tag = t.tag if not prefix_strip else t.tag.rsplit('}', 1)[-1]
    d = {tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            #            print(dir(dc))
            for k, v in dc.items():
                if prefix_strip:
                    k = k.rsplit('}', 1)[-1]
                dd[k].append(v)
        d = {tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[tag].update(('@' + k.rsplit('}', 1)[-1], v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            tag = tag.rsplit('}', 1)[-1]
            if text:
                d[tag]['#text'] = text
        else:
            d[tag] = text
    return d


def xml_to_jsonl(fromname, toname, options={}, default_options={'prefix_strip': True}):
    options = __copy_options(options, default_options)
    ins = open(fromname, 'rb')  # , encoding='utf-8')
    outf = open(toname, 'wb')
    n = 0
    for event, elem in etree.iterparse(ins):
        shorttag = elem.tag.rsplit('}', 1)[-1]
        if shorttag == options['tagname']:
            n += 1
            if options['prefix_strip']:
                j = etree_to_dict(elem, prefix_strip=options['prefix_strip'])
            else:
                j = etree_to_dict(elem)
            outf.write(orjson.dumps(j[shorttag]))
            outf.write(LINEEND)
        if n % 500 == 0:
            logging.info('xml2jsonl: processed %d xml tags' % (n))
    logging.info('xml2jsonl: processed %d xml tags finally' % (n))
    ins.close()
    outf.close()


def xls_to_csv(fromname, toname, options={},
               default_options={'start_line': 0, 'skip_end_rows': 0, 'delimiter': ',', 'encoding': 'utf8'}):
    options = __copy_options(options, default_options)
    b = load_xls(fromname)
    s = b.sheet_by_index(0)
    bc = open(toname, 'w', encoding=options['encoding'])
    bcw = csv.writer(bc, delimiter=options['delimiter'])
    n = 0
    for row in range(options['start_line'], s.nrows - options['skip_end_rows']):
        n += 1
        this_row = []
        for col in range(s.ncols):
            v = str(s.cell_value(row, col))
            v = v.replace('\n', ' ').strip()
            #			v = v.encode('utf8') if type(v) == type(u'') else str(v)
            this_row.append(v)
        bcw.writerow(this_row)
        if n % 10000 == 0:
            logging.info('xls2csv: processed %d records' % (n))
    bc.close()


def csv_to_bson(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ','}):
    options = __copy_options(options, default_options)
    source = open(fromname, 'r', encoding=options['encoding'])
    output = open(toname, 'wb')
    reader = csv.DictReader(source, delimiter=options['delimiter'])
    n = 0
    for j in reader:
        n += 1
        rec = bson.BSON.encode(j)
        output.write(rec)
        if n % 10000 == 0:
            logging.info('csv2bson: processed %d records' % (n))
    source.close()
    output.close()


def csv_to_jsonl(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ','}):
    options = __copy_options(options, default_options)
    source = open(fromname, 'r', encoding=options['encoding'])
    output = open(toname, 'wb')
    reader = csv.DictReader(source, delimiter=options['delimiter'])
    n = 0
    for j in reader:
        n += 1
        output.write(json.dumps(j, ensure_ascii=False).encode('utf8'))
        #        output.write(orjson.dumps(j, ensure_ascii=False).encode('utf8', ))
        output.write(u'\n'.encode('utf8'))
        if n % 10000 == 0:
            logging.info('csv2jsonl: processed %d records' % (n))
    source.close()
    output.close()


def xls_to_jsonl(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0, 'fields': None}):
    options = __copy_options(options, default_options)
    source = load_xls(fromname)
    output = open(toname, 'wb')
    sheet = source.sheet_by_index(options['start_page'])
    n = 0
    fields = options['fields'].split(',') if options['fields'] is not None else None
    for rownum in range(options['start_line'], sheet.nrows):
        n += 1
        tmp = list()
        for i in range(0, sheet.ncols):
            tmp.append(sheet.row_values(rownum)[i])
        if n == 1 and fields is None:
            fields = tmp
            continue
        line = orjson.dumps(dict(zip(fields, tmp)))
        output.write(line + LINEEND)
        #        output.write(u'\n'.encode('utf8'))
        if n % 10000 == 0:
            logging.info('xls2jsonl: processed %d records' % (n))
    output.close()




def xlsx_to_jsonl(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0}):
    from openpyxl import load_workbook as load_xlsx
    options = __copy_options(options, default_options)
    source = load_xlsx(fromname)
    output = open(toname, 'wb')
    sheet = source.active       #FIXME! Use start_page instead
    n = 0
    fields = options['fields'].split(',') if options['fields'] is not None else None
    for row in sheet.iter_rows():
        n += 1
        if n < options['start_line']: 
            continue
        tmp = list()

        for cell in row:
            tmp.append(cell.value)
        if n == 1 and fields is None:
            fields = tmp
            continue
        line = orjson.dumps(dict(zip(fields, tmp)))
        output.write(line)
        output.write(LINEEND)
        if n % 10000 == 0:
            logging.debug('xlsx2jsonl: processed %d records' % (n))
    source.close
    output.close()

def xlsx_to_bson(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0}):
    from openpyxl import load_workbook as load_xlsx
    options = __copy_options(options, default_options)
    source = load_xlsx(fromname)
    output = open(toname, 'wb')
    sheet = source.active       #FIXME! Use start_page instead
    n = 0
    fields = options['fields'].split(',') if options['fields'] is not None else None
    for row in sheet.iter_rows():
        n += 1
        if n < options['start_line']: 
            continue
        tmp = list()

        for cell in row:
            tmp.append(cell.value)
        if n == 1 and fields is None:
            fields = tmp
            continue
        output.write(bson.BSON.encode(dict(zip(fields, tmp))))

        if n % 10000 == 0:
            logging.debug('xlsx2bson: processed %d records' % (n))
    source.close
    output.close()

def xls_to_bson(fromname, toname, options={}, default_options={'start_page': 0, 'start_line': 0}):
    options = __copy_options(options, default_options)
    source = load_xls(fromname)
    output = open(toname, 'wb')
    sheet = source.sheet_by_index(options['start_page'])
    n = 0
    for rownum in range(options['start_line'], sheet.nrows):
        n += 1
        tmp = list()
        for i in range(0, sheet.ncols):
            tmp.append(sheet.row_values(rownum)[i])
        output.write(bson.BSON.encode(dict(zip(options['fields'], tmp))))
        if n % 10000 == 0:
            logging.info('xls2bson: processed %d records' % (n))
    source.close
    output.close()


def _is_flat(item):
    for k, v in item.items():
        if isinstance(v, dict) or isinstance(v, tuple) or isinstance(v, list):
            return False
    return True


def express_analyze_jsonl(filename, itemlimit=100):
    f = open(filename, 'r', encoding='utf8')
    isflat = True
    n = 0
    keys = set()
    for line in f:
        n += 1
        if n > itemlimit: 
            break
        record = orjson.loads(line)
        if isflat:
            if not _is_flat(record):
                isflat = False
        if len(keys) == 0:
            keys = set(record.keys())
        else:
            keys = keys.union(set(record.keys()))
    f.close()
    keys = list(keys)
    keys.sort()
    return {'isflat': isflat, 'keys': keys}


def jsonl_to_csv(fromname, toname, options={},
                 default_options={'force_flat': False, 'useitems': 100, 'delimiter': ','}):
    options = __copy_options(options, default_options)
    analysis = express_analyze_jsonl(fromname, itemlimit=options['useitems'])
    if not options['force_flat'] and not analysis['isflat']:
        logging.error("File %s is not flat and 'force_flat' flag not set. File not converted")
        return
    out = open(toname, 'w', encoding='utf8')
    writer = csv.writer(out, delimiter=options['delimiter'])
    writer.writerow(analysis['keys'])
    f = open(fromname, 'r', encoding='utf8')
    keys = analysis['keys']
    n = 0
    for line in f:
        n += 1
        record = orjson.loads(line)
        item = []
        for k in keys:
            if k in record.keys():
                item.append(record[k])
            else:
                item.append('')
        writer.writerow(item)
        if n % 10000 == 0:
            logging.info('jsonl2csv: processed %d records' % (n))
    out.close()
    f.close()
    pass


def default(obj):
    if isinstance(obj, ObjectId):
        return str(obj)

def bson_to_jsonl(fromname, toname, options={}, default_options={}):
    options = __copy_options(options, default_options)
    source = open(fromname, 'rb')
    output = open(toname, 'wb')
    n = 0
    for r in bson.decode_file_iter(source):
        n += 1
        output.write(orjson.dumps(r, default=default))
        output.write(LINEEND)
        if n % 10000 == 0:
            logging.info('bson2jsonl: processed %d records' % (n))
    source.close()
    output.close()


def json_to_jsonl(fromname, toname, options={}, default_options={}):
    """Simple implementation of JSON to JSON lines conversion. Assuming that JSON is an array or dict with 1-st level value with data"""
    options = __copy_options(options, default_options)    
    source = open(fromname, 'rb')
    source_data = json.load(source)    
    data = source_data
    if 'tagname' in options.keys():
        if isinstance(source_data, dict) and  options['tagname'] in source_data.keys():
            data = data[options['tagname']]    
    output = open(toname, 'wb')
    n = 0
    for r in data:
        n += 1
        output.write(orjson.dumps(r) + LINEEND)
        if n % 10000 == 0:
            logging.info('json2jsonl: processed %d records' % (n))
    source.close()
    output.close()


def csv_to_parquet(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ',', 'compression' : 'brotli'}):
    options = __copy_options(options, default_options)
    df = pandas.read_csv(fromname, delimiter=options['delimiter'], encoding=options['encoding'])
    df.to_parquet(toname, compression=options['compression'] if options['compression'] != 'None' else None)


def jsonl_to_parquet(fromname, toname, options={},
                 default_options={'force_flat': False, 'useitems': 100, 'compression' : 'brotli'}):                 
    options = __copy_options(options, default_options)
    df = pandas.read_json(fromname, lines=True, encoding=options['encoding'])
    df.to_parquet(toname, compression=options['compression'] if options['compression'] != 'None' else None)


PYORC_COMPRESSION_MAP = {'zstd': 5, 'snappy' : 2, 'zlib' : 1, 'lzo' : 3, 'lz4' : 4, 'None' : 0}

def csv_to_orc(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ',', 'compression' : 'zstd'}):
    """Converts CSV file to ORC file"""
    import pyorc    
    options = __copy_options(options, default_options)
    compression = PYORC_COMPRESSION_MAP[options['compression']] if options['compression'] in PYORC_COMPRESSION_MAP.keys() else 0
    source = open(fromname, 'r', encoding=options['encoding'])
    reader = csv.DictReader(source, delimiter=options['delimiter'])
    struct_schema = []
    for field in reader.fieldnames:
        struct_schema.append('%s:string' %(field))
    output = open(toname, 'wb')
    writer = pyorc.Writer(output, "struct<%s>" % (','.join(struct_schema)), struct_repr = pyorc.StructRepr.DICT, compression=compression, compression_strategy=1)
    n = 0
    for row in reader:
        n += 1
        try: 
            writer.write(row)
        except TypeError:
            print('Error processing row %d. Skip and continue' % (n))
    writer.close()
    output.close()

def jsonl_to_orc(fromname, toname, options={},
                 default_options={'force_flat': False, 'useitems': 100, 'compression' : 'zstd'}):
    """Converts JSON file to ORC file"""
    import pyorc
    options = __copy_options(options, default_options)
    compression = PYORC_COMPRESSION_MAP[options['compression']] if options['compression'] in PYORC_COMPRESSION_MAP.keys() else 0
    df = pandas.read_json(fromname, lines=True, encoding=options['encoding'])
    df.info()
    struct_schema = df_to_pyorc_schema(df)
    output = open(toname, 'wb')
    writer = pyorc.Writer(output, "struct<%s>" % (','.join(struct_schema)), struct_repr = pyorc.StructRepr.DICT, compression=compression, compression_strategy=1)
    writer.writerows(df.to_dict(orient="records"))
    writer.close()
    output.close()

def csv_to_avro(fromname, toname, options={}, default_options={'encoding': 'utf8', 'delimiter': ',', 'compression' : 'deflate'}):
    """Converts CSV file to AVRO file"""
    import avro.schema
    from avro.datafile import DataFileWriter
    from avro.io import DatumWriter
    
    options = __copy_options(options, default_options)    
    options['compression']
    source = open(fromname, 'r', encoding=options['encoding'])
    reader = csv.DictReader(source, delimiter=options['delimiter'])

    schema_dict = {"namespace": "data.avro", "type": "record", "name": "Record", "fields": []}

    for field in reader.fieldnames:
        schema_dict['fields'].append({'name' : field, 'type' : 'string'})
    schema = avro.schema.parse(json.dumps(schema_dict))
    output = open(toname, 'wb')
    writer = DataFileWriter(output, DatumWriter(), schema, codec=options['compression'])
    n = 0
    for row in reader:
        n += 1
        try: 
            writer.append(row)
        except TypeError:
            print('Error processing row %d. Skip and continue' % (n))
    writer.close()
    output.close()

CONVERT_FUNC_MAP = {
    'xls2csv': xls_to_csv,
    'xls2jsonl': xls_to_jsonl,
    'xls2bson': xls_to_bson,
    'xlsx2jsonl': xlsx_to_jsonl,
    'xlsx2bson': xlsx_to_bson,
    'csv2jsonl': csv_to_jsonl,
    'csv2bson': csv_to_bson,
    'xml2jsonl': xml_to_jsonl,
    'jsonl2csv': jsonl_to_csv,
    'bson2jsonl': bson_to_jsonl,
    'json2jsonl': json_to_jsonl,
    'csv2parquet' : csv_to_parquet,
    'jsonl2parquet': jsonl_to_parquet,
    'jsonl2orc' : jsonl_to_orc,
    'csv2orc' : csv_to_orc,
    'csv2avro' : csv_to_avro,
}


DEFAULT_HEADERS_DETECT_LIMIT = 1000

def make_flat(item):
    result = {}
    for k, v in item.items():
        if isinstance(v, tuple) or isinstance(v, list) or isinstance(v, dict):
            result[k] = str(v)
        result[k] = v
    return result

class Converter:
    def __init__(self, batch_size = DEFAULT_BATCH_SIZE):
        self.batch_size = batch_size
        pass

    def convert(self, fromfile, tofile, options={}, limit=DEFAULT_HEADERS_DETECT_LIMIT):
        iterableargs = get_iterable_options(options)
#        print(iterableargs)
        it_in = open_iterable(fromfile, mode='r', iterableargs=iterableargs)       
        is_flatten = get_option(options, 'flatten')
        keys = []
        n = 0
        logging.info('Extracting schema')
        for item in tqdm(it_in, total=limit):
#            print(item)
            if limit is not None and n > limit:
                break
            n += 1                
            if not is_flatten:
                dk = dict_generator(item)
                for i in dk:
                    k = ".".join(i[:-1])
                    if k not in keys:
                        keys.append(k)
            else:
                item = make_flat(item)
                for k in item.keys():
                    if k not in keys:
                        keys.append(k)

        it_in.reset()
        it_out = open_iterable(tofile, mode='w', iterableargs={'keys' : keys})

        logging.info('Converting data')
        n = 0
        batch = []
        for row in tqdm(it_in):
            n += 1 
            if is_flatten:
                for k in keys:
                    if k not in row.keys(): 
                        row[k] = None              
                batch.append(make_flat(row))
            else:
                batch.append(row)
            if n % self.batch_size == 0:
                it_out.write_bulk(batch)
                batch = []
        if len(batch) > 0: 
            it_out.write_bulk(batch)
        it_in.close()
        it_out.close()


    def convert_old(self, fromfile, tofile, options={}):
        fromtype = options['format_in'] if options['format_in'] is not None else get_file_type(fromfile)
        totype = options['format_out'] if options['format_out'] is not None else get_file_type(tofile)
        key = '%s2%s' % (fromtype, totype)
        func = CONVERT_FUNC_MAP.get(key, None)
        if func is None:
            logging.error('Conversion between %s and %s not supported' % (fromtype, totype))
        else:
            logging.info('Convert %s from %s to %s' % (key, fromfile, tofile))
            func(fromfile, tofile, options)



# -*- coding: utf8 -*-
from collections import OrderedDict
import chardet
from .constants import SUPPORTED_FILE_TYPES
from .constants import DEFAULT_OPTIONS

def detect_encoding(filename, limit=1000000):
    f = open(filename, 'rb')
    chunk = f.read(limit)
    f.close()
    detected = chardet.detect(chunk)
    return detected

def detect_delimiter(filename, encoding='utf8'):
    f = open(filename, 'r', encoding=encoding)
    line = f.readline()
    f.close()
    dict1 = {',': line.count(','), ';': line.count(';'), '\t': line.count('\t'), '|' : line.count('|')}
    delimiter = max(dict1, key=dict1.get)
    return delimiter

def get_file_type(filename):
    ext = filename.rsplit('.', 1)[-1].lower()
    if ext in SUPPORTED_FILE_TYPES:
        return ext
    return None

def get_option(options, name):
    """Returns value of the option"""
    if name in options.keys():
        return options[name]
    elif name in DEFAULT_OPTIONS.keys():
        return DEFAULT_OPTIONS[name]
    return None

def get_dict_value(d, keys):
    out = []
    if d is None:
        return out
#    keys = key.split('.')
    if len(keys) == 1:
        if type(d) == type({}) or isinstance(d, OrderedDict):
            if keys[0] in d.keys():
                out.append(d[keys[0]])
        else:
            for r in d:
                if r and keys[0] in r.keys():
                    out.append(r[keys[0]])
#        return out
    else:
        if type(d) == type({}) or isinstance(d, OrderedDict):
            if keys[0] in d.keys():
                out.extend(get_dict_value(d[keys[0]], keys[1:]))
        else:
            for r in d:
                if keys[0] in r.keys():
                    out.extend(get_dict_value(r[keys[0]], keys[1:]))
    return out


def strip_dict_fields(record, fields, startkey=0):
    keys = record.keys()
    localf = []
    for field in fields:
        if len(field) > startkey:
            localf.append(field[startkey])
    for k in list(keys):
        if k not in localf:
            del record[k]

    if len(k) > 0:
        for k in record.keys():
            if type(record[k]) == type({}):
                record[k] = strip_dict_fields(record[k], fields, startkey + 1)
    return record


def dict_generator(indict, pre=None):
    """Processes python dictionary and return list of key values
    :param indict
    :param pre
    :return generator"""
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in list(indict.items()):
            if key == "_id":
                continue
            if isinstance(value, dict):
                #                print 'dgen', value, key, pre
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    if isinstance(v, dict):
                        #                print 'dgen', value, key, pre
                        for d in dict_generator(v, pre + [key]):
                            yield d
#                    for d in dict_generator(v, [key] + pre):
#                        yield d
            else:
                yield pre + [key, value]
    else:
        yield indict


def guess_int_size(i):
    if i < 255:
        return 'uint8'
    if i < 65535:
        return 'uint16'
    return 'uint32'

def guess_datatype(s, qd):
    """Guesses type of data by string provided
    :param s
    :param qd
    :return datatype"""
    attrs = {'base' : 'str'}
#    s = unicode(s)
    if s is None:
       return {'base' : 'empty'}
    if type(s) == type(1):
        return {'base' : 'int'}
    if type(s) == type(1.0):
        return {'base' : 'float'}
    elif type(s) != type(''):
#        print((type(s)))
        return {'base' : 'typed'}
#    s = s.decode('utf8', 'ignore')
    if s.isdigit():
        if s[0] == 0:
            attrs = {'base' : 'numstr'}
        else:
            attrs = {'base' : 'int', 'subtype' : guess_int_size(int(s))}
    else:
        try:
            float(s)
            attrs = {'base' : 'float'}
            return attrs
        except ValueError:
            pass
        if qd:
            is_date = False
            res = qd.match(s)
            if res:
                attrs = {'base': 'date', 'pat': res['pattern']}
                is_date = True
            if not is_date:
                if len(s.strip()) == 0:
                    attrs = {'base' : 'empty'}
    return attrs


def buf_count_newlines_gen(fname):
    def _make_gen(reader):
        while True:
            b = reader(2 ** 16)
            if not b: break
            yield b

    with open(fname, "rb") as f:
        count = sum(buf.count(b"\n") for buf in _make_gen(f.raw.read))
    return count


def get_dict_keys(iterable, limit=1000):
    n = 0
    keys = []
    for item in iterable:
        if limit and n > limit:
            break
        n += 1
        dk = dict_generator(item)
        for i in dk:
            k = ".".join(i[:-1])
            if k not in keys:
                keys.append(k)
    return keys


def _is_flat(item):
    """Measures if object is flat"""
    for k, v in item.items():
        if isinstance(v, tuple) or isinstance(v, list):
            return False
        elif isinstance(v, dict):
            if not _is_flat(v): return False
    return True



# -*- coding: utf8 -*-
import os
import csv
# import json
import logging
import sys
import zipfile

import bson
import dictquery as dq
import orjson

# from xmlr import xmliter
from ..utils import get_file_type, get_option, get_dict_value, strip_dict_fields, dict_generator, detect_encoding

from iterable.helpers.detect import open_iterable

from ..common.iterable import IterableData, DataWriter
LINEEND = u'\n'.encode('utf8')

ITERABLE_OPTIONS_KEYS = ['tagname', 'delimiter', 'encoding', 'start_line', 'page']


def get_iterable_options(options):
    out = {}
    for k in ITERABLE_OPTIONS_KEYS:
        if k in options.keys():
            out[k] = options[k]
    return out            


def get_iterable_fields_uniq(iterable, fields, dolog=False, dq=None):
    """ Returns all uniq values of the fields of iterable dictionary"""
    n = 0
    uniqval = []
    for row in iterable:
#        print(row)
        n += 1
        if dolog and n % 1000 == 0:
            logging.debug('uniq: processing %d records' % (n))
        try:
            allvals = []
            for field in fields:
                allvals.append(get_dict_value(row, field.split('.')))

            for n1 in range(0, len(allvals[0]), 1):
                k = []
                for n2 in range(0, len(allvals)):
                    k.append(str(allvals[n2][n1]))
                if k not in uniqval:
                    uniqval.append(k)
        except KeyError:
            pass
    return uniqval

def get_iterable_fields_freq(iterable, fields, dolog=False, filter=None, dq=None):
    """Iterates and returns most frequent values"""
    n = 0
    valuedict = {}
    for r in iterable:
        n += 1
        if dolog and n % 10000 == 0:
            logging.info('frequency: processing %d records' % (n))
        if filter is not None:
            if not dq.match(r, filter):
                continue
        try:
            allvals = []
            for field in fields:
                allvals.append(get_dict_value(r, field.split('.')))

            for n1 in range(0, len(allvals[0]), 1):
                k = []
                for n2 in range(0, len(allvals)):
                    k.append(str(allvals[n2][n1]))
                kx = '\t'.join(k)
                v = valuedict.get(kx, 0)
                valuedict[kx] = v + 1
        except KeyError:
            pass
    return valuedict


class Selector:
    def __init__(self):
        pass

    def uniq(self, fromfile, options={}):
        """Extracts unique values by field"""
        logging.debug('Processing %s' % fromfile)
        iterableargs = get_iterable_options(options)
        iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)
        to_file = get_option(options, 'output')
        if to_file:
            to_type = get_file_type(to_file)
            if not to_file:
                logging.debug('Output file type not supported')
                return
            out = open(to_file, 'w', encoding='utf8')
        else:
            to_type = 'csv'
            out = sys.stdout
        fields = options['fields'].split(',')
        logging.info('uniq: looking for fields: %s' % (options['fields']))
        uniqval = get_iterable_fields_uniq(iterable, fields, dolog=True)
        iterable.close()
        logging.debug('%d unique values found' % (len(uniqval)))
        writer = DataWriter(out, filetype=to_type, fieldnames=fields)
        writer.write_items(uniqval)


    def headers(self, fromfile, options={}):
        """Extracts headers values"""
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        limit = get_option(options, 'limit')
        iterableargs = get_iterable_options(options)
        
        iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)
        keys = []
        n = 0
        for item in iterable:
            if limit and n > limit:
                break
            n += 1
            dk = dict_generator(item)
            for i in dk:
                k = ".".join(i[:-1])
                if k not in keys:
                    keys.append(k)
        iterable.close()
        output = get_option(options, 'output')
        if output:
            f = open(output, 'w', encoding=get_option(options, 'encoding'))
            f.write('\n'.join(keys))
            f.close()
        else:
            for x in keys:
                print(x.encode('utf8').decode('utf8', 'ignore'))  

    def frequency(self, fromfile, options={}):
        """Calculates frequency of the values in the file"""
        iterableargs = get_iterable_options(options)
        iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)
        to_file = get_option(options, 'output')
        if to_file:
            get_file_type(to_file)
            if not to_file:
                print('Output file type not supported')
                return
            open(to_file, 'w', encoding='utf8')
        else:
            pass
        fields = options['fields'].split(',')
        valuedict = {}
        if iterable:
            valuedict = get_iterable_fields_freq(iterable, fields, dolog=True)
        else:
            logging.info('File type not supported')
            return
        logging.debug('frequency: %d unique values found' % (len(valuedict)))
        thedict = sorted(valuedict.items(), key=lambda item: item[1], reverse=False)
        output = get_option(options, 'output')
        strkeys = '\t'.join(fields) + '\tcount'
        if output:
            encoding = get_option(options, 'encoding')
            f = open(output, 'wb')
            f.write((strkeys + '\n').encode(encoding))
            for k, v in thedict:
                f.write(('%s\t%d\n' % (k, v)).encode(encoding))
            f.close()
        else:
            print(strkeys)
            for k, v in thedict:
                print('%s\t%d' % (k, v))

    def select(self, fromfile, options={}):
        """Select or re-order columns from file"""
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        iterable = IterableData(fromfile, options=options)
        to_file = get_option(options, 'output')

        if to_file:
            to_type = get_file_type(to_file)
            if not to_file:
                print('Output file type not supported')
                return
            if to_type == 'bson':
                out = open(to_file, 'wb')
            if to_type == 'jsonl':
                out = open(to_file, 'wb')
            else:
                out = open(to_file, 'w', encoding='utf8')
        else:
            to_type = f_type
            out = sys.stdout
        fields = options['fields'].split(',')
        writer = DataWriter(out, filetype=to_type, fieldnames=fields)
        if iterable:
            n = 0
            fields = [field.split('.') for field in fields]
            chunk = []
            for r in iterable.iter():
                n += 1
                if options['filter'] is not None:
                    res = dq.match(r, options['filter'])
                    #                    print(options['filter'], r)
                    if not res:
                        continue
                r_selected = strip_dict_fields(r, fields, 0)
                if n % 1000 == 0:
                    logging.info('select: processing %d records of %s' % (n, fromfile))
                    if len(chunk) > 0:
                        writer.write_items(chunk)
                        chunk = []
                else:
                    chunk.append(r_selected)
            if len(chunk) > 0:
                writer.write_items(chunk)
        else:
            logging.info('File type not supported')
            return
        logging.debug('select: %d records processed' % (n))
        out.close()


    def split_new(self, fromfile, options={}):
        """Splits the given file with data into chunks based on chunk size or field value"""
        iterableargs = get_iterable_options(options)
        iterable = open_iterable(fromfile, mode='r', iterableargs=iterableargs)
        to_file = get_option(options, 'output')


    def split(self, fromfile, options={}):
        """Splits the given file with data into chunks based on chunk size or field value"""
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        if options['zipfile']:
            z = zipfile.ZipFile(fromfile, mode='r')
            fnames = z.namelist()
            finfilename = fnames[0]
            if f_type == 'bson':
                infile = z.open(fnames[0], 'rb')
            else:
                infile = z.open(fnames[0], 'r')
        elif options['gzipfile']:
            import gzip
            infile = gzip.open(fromfile, 'rb')
            finfilename = fromfile.split('.', 1)[0] + '.' + f_type
        else:
            finfilename = fromfile
            if f_type == 'bson':
                infile = open(fromfile, 'rb')
            else:
                if 'encoding' in options.keys():
                    infile = open(fromfile, 'r', encoding=get_option(options, 'encoding'))
                else:
                    detected_enc = detect_encoding(fromfile, limit=100000)
                    if detected_enc:
                        infile = open(fromfile, 'r', encoding=detected_enc['encoding'])
                    else:
                        infile = open(fromfile, 'r', encoding='utf8')
        fields = options['fields'].split(',') if options['fields'] is not None else None
        valuedict = {}
        delimiter = get_option(options, 'delimiter')
        if f_type == 'csv':
            reader = csv.DictReader(infile, delimiter=delimiter)
            n = 0
            chunknum = 1
            if options['fields'] is None:
                splitname = finfilename.rsplit('.', 1)[0] + '_%d.csv' % (chunknum)
                out = open(splitname, 'w', encoding=get_option(options, 'encoding'))
                writer = csv.DictWriter(out, fieldnames=reader.fieldnames, delimiter=delimiter)
                writer.writeheader()
                for r in reader:
                    n += 1
                    if n % 10000 == 0:
                        logging.info('split: processing %d records of %s' % (n, fromfile))
                    if options['filter'] is not None:
                        if not dq.match(r, options['filter']):
                            continue
                    writer.writerow(r)
                    if n % options['chunksize'] == 0:
                        out.close()
                        chunknum += 1
                        splitname = finfilename.rsplit('.', 1)[0] + '_%d.csv' % (chunknum)
                        out = open(splitname, 'w', encoding=get_option(options, 'encoding'))
                        writer = csv.DictWriter(out, fieldnames=reader.fieldnames, delimiter=delimiter)
                        writer.writeheader()
        elif f_type == 'jsonl':
            n = 0
            chunknum = 1
            if options['fields'] is None:
                splitname = finfilename.rsplit('.', 1)[0] + '_%d.jsonl' % (chunknum)
                out = open(splitname, 'wb')  # , encoding=get_option(options, 'encoding'))

                for l in infile:
                    n += 1
                    if n % 10000 == 0:
                        logging.info('split: processing %d records of %s' % (n, fromfile))
                    r = orjson.loads(l)
                    if options['filter'] is not None:
                        if not dq.match(r, options['filter']):
                            continue
                    out.write(orjson.dumps(r, option=orjson.OPT_APPEND_NEWLINE))
                    if n % options['chunksize'] == 0:
                        out.close()
                        chunknum += 1
                        splitname = finfilename.rsplit('.', 1)[0] + '_%d.jsonl' % (chunknum)
                        logging.info('split: new chunk %s' % splitname)
                        out = open(splitname, 'wb') #, encoding=get_option(options, 'encoding'))
            else:
                for l in infile:
                    n += 1
                    if n % 10000 == 0:
                        logging.info('split: processing %d records of %s' % (n, fromfile))
                    r = orjson.loads(l)
                    if options['filter'] is not None:
                        if not dq.match(r, options['filter']):
                            continue
                    try:
                        kx = get_dict_value(r, fields[0].split('.'))[0]
                    except IndexError:
                        continue
                        kx = "None"
                    if kx is None: continue
                    kx = kx.replace('\\', '-').replace('/', '-').replace('?', '-').replace('<', '-').replace('>', '-').replace('\n', '')
                    v = valuedict.get(kx, None)
                    if v is None:
#                        splitname = finfilename.rsplit('.', 1)[0] + '_%s.jsonl' % (kx)
                        splitname = '%s.jsonl' % (kx)
                        if options['dirname'] is not None: splitname = os.path.join(options['dirname'], splitname)
                        valuedict[kx] = open(splitname, 'w', encoding='utf8')
                    valuedict[kx].write(l)
                #                    valuedict[kx].write(l.decode('utf8'))#.decode('utf8')#)
                for opened in valuedict.values():
                    opened.close()
        elif f_type == 'bson':
            bson_iter = bson.decode_file_iter(infile)
            n = 0
            for r in bson_iter:
                n += 1
                #                print(r)
                strip_dict_fields(r, fields, 0)
                #                out.write(json.dumps(r_selected)+'\n')
                if n % 10000 == 0:
                    logging.info('split: processing %d records of %s' % (n, fromfile))

        else:
            logging.info('File type not supported')
            return
        logging.debug('split: %d records processed' % (n))



# -*- coding: utf8 -*-
import csv
import zipfile
import sys
import orjson
import bson
import logging
#from xmlr import xmliter
from ..utils import get_file_type, get_option, get_dict_value
import dictquery as dq
from ..validate import VALIDATION_RULEMAP


class Validator:
    def __init__(self):
        pass

    def validate(self, fromfile, options={}):
        """Validates selected field against validation rule"""
        logging.debug('Processing %s' % fromfile)
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        if options['zipfile']:
            z = zipfile.ZipFile(fromfile, mode='r')
            fnames = z.namelist()
            if f_type == 'bson':
                infile = z.open(fnames[0], 'rb')
            else:
                infile = z.open(fnames[0], 'r')
        else:
            if f_type == 'bson':
                infile = open(fromfile, 'rb')
            else:
                infile = open(fromfile, 'r', encoding=get_option(options, 'encoding'))
        to_file = get_option(options, 'output')
        if to_file:
            get_file_type(to_file)
            if not to_file:
                logging.debug('Output file type not supported')
                return
            out = open(to_file, 'w', encoding='utf8')
        else:
            out = sys.stdout
        fields = options['fields'].split(',')
        val_func = VALIDATION_RULEMAP[options['rule']]
        logging.info('uniq: looking for fields: %s' % (options['fields']))
        validated = []
        stats = {'total': 0, 'invalid': 0, 'novalue' : 0}
        if f_type == 'csv':
            delimiter = get_option(options, 'delimiter')
            reader = csv.DictReader(infile, delimiter=delimiter)
            n = 0
            for r in reader:
                n += 1
                if n % 1000 == 0:
                    logging.info('uniq: processing %d records of %s' % (n, fromfile))
                if options['filter'] is not None:
                    if not dq.match(r, options['filter']):
                        continue
                res = val_func(r[fields[0]])
                stats['total'] += 1
                if not res:
                    stats['invalid'] += 1
                validated.append({fields[0] : r[fields[0]], fields[0] + '_valid' : res})

        elif f_type == 'jsonl':
            n = 0
            for l in infile:
                n += 1
                if n % 10000 == 0:
                    logging.info('uniq: processing %d records of %s' % (n, fromfile))
                r = orjson.loads(l)
                if options['filter'] is not None:
                    if not dq.match(r, options['filter']):
                        continue
                stats['total'] += 1
                values = get_dict_value(r, fields[0].split('.'))
                if len(values) > 0:
                    res = val_func(values[0])
                    if not res:
                        stats['invalid'] += 1
                    validated.append({fields[0] : values[0], fields[0] + '_valid' : res})
                else:
                    stats['novalue'] += 1

        elif f_type == 'bson':
            bson_iter = bson.decode_file_iter(infile)
            n = 0
            for r in bson_iter:
                n += 1
                if n % 1000 == 0:
                    logging.info('uniq: processing %d records of %s' % (n, fromfile))
                if options['filter'] is not None:
                    if not dq.match(r, options['filter']):
                        continue
                stats['total'] += 1
                values = get_dict_value(r, fields[0].split('.'))
                if len(values) > 0:
                    res = val_func(values[0])
                    if not res:
                        stats['invalid'] += 1
                    validated.append({fields[0] : values[0], fields[0] + '_valid' : res})
                else:
                    stats['novalue'] += 1
        else:
            logging.error('Invalid filed format provided')
            return
        infile.close()
        stats['share'] = 100.0 * stats['invalid'] / stats['total']
        logging.debug('validate: complete, %d records (%.2f%%) not valid and %d (%.2f%%) not found of %d against %s' % (stats['invalid'], stats['share'], stats['novalue'], 100.0 * stats['novalue'] / stats['total'], stats['total'], options['rule']))
        if options['mode'] != 'stats':
            writer = csv.DictWriter(out, fieldnames=[fields[0], fields[0] + '_valid'], delimiter=get_option(options, 'delimiter'))
            for row in validated:
                if options['mode'] == 'invalid':
                    if not row[fields[0] + '_valid']:
                        writer.writerow(row)
                elif options['mode'] == 'all':
                    writer.writerow(row)
        else:
            out.write(str(orjson.dumps(stats, option=orjson.OPT_INDENT_2)))




# -*- coding: utf8 -*-
#!/usr/bin/env python
"""The main entry point. Invoke as `undatum' or `python -m undatum`.

"""
import sys


def main():
    try:
        from .core import app
        app()
    except KeyboardInterrupt:
        print("Ctrl-C pressed. Aborting")
    sys.exit(0)


if __name__ == '__main__':
    app()



# -*- coding: utf8 -*-

import csv
import xlwt
import openpyxl
import json
import datetime

from docx import Document
from docx.table import _Cell
from docx.oxml.simpletypes import ST_Merge


def __extract_table(table, strip_space=False):
    """Extracts table data from table object"""
    results = []
    n = 0
    for tr in table._tbl.tr_lst:
        r = []
        for tc in tr.tc_lst:
            for grid_span_idx in range(tc.grid_span):
                if tc.vMerge == ST_Merge.CONTINUE:
                    value = results[n - 1][len(r) - 1]
                elif grid_span_idx > 0:
                    value = r[-1]
                else:
                    cell = _Cell(tc, table)
                    value = cell.text.replace("\n", " ")
                if strip_space:
                    value = value.strip()
                r.append(value)
        results.append(r)
#        print(r)
        n += 1
    return results


def __store_table(tabdata, filename, format="csv"):
    """Saves table data as csv file"""
    if format == "csv":
        f = open(filename, "w", encoding='utf8')
        w = csv.writer(f, delimiter=",")
        for row in tabdata:
            w.writerow(row)
    elif format == 'tsv':
        f = open(filename, 'w')
        w = csv.writer(f, delimiter='\t')
        for row in tabdata:
            w.writerow(row)
    elif format == 'xls':
        workbook = xlwt.Workbook()
        ws = __xls_table_to_sheet(tabdata, workbook.add_sheet("0"))
#        print(dir(ws))
        workbook.save(filename)
    elif format == "xlsx":
        workbook = openpyxl.Workbook()
        ws = __xlsx_table_to_sheet(tabdata, workbook.create_sheet("0"))
        workbook.save(filename)

def __xls_table_to_sheet(table, ws):
    rn = 0
    for row in table:
        cn = 0
        for c in row:
            ws.write(rn, cn, c)
            cn += 1
        rn += 1
    return ws


def __xlsx_table_to_sheet(table, ws):
    rn = 0
    for row in table:
        ws.append(row)
        rn += 1
    return ws


def extract_docx_tables(filename, strip_space=True):
    """Extracts table from .DOCX files"""
    tables = []
    document = Document(filename)
    n = 0
    for table in document.tables:
        n += 1
        info = {}
        info['id'] = n
        info['num_cols'] = len(table.columns)
        info['num_rows'] = len(table.rows)
        info['style'] = table.style.name
        tdata = __extract_table(table, strip_space=strip_space)
        info['data'] = tdata
        tables.append(info)
    return tables





def extract(filename, format="csv", sizefilter=0, singlefile=False, output=None, strip_space=True):
    """Extracts tables from csv files and saves them as csv, xls or xlsx files"""
    tables = extract_tables(filename, strip_space=strip_space)
    name = filename.rsplit(".", 1)[0]
    format = format.lower()
    n = 0
    lfilter = int(sizefilter)
    if singlefile:
        if format == "xls":
            workbook = xlwt.Workbook()
            for t in tables:
                if lfilter >= len(t):
                    continue
                n += 1
                ws = __xls_table_to_sheet(t['data'], workbook.add_sheet(str(n)))
            destname = output if output else name + ".%s" % (format)
            workbook.save(destname)
        elif format == "xlsx":
            workbook = openpyxl.Workbook()
            for t in tables:
                if lfilter >= len(t):
                    continue
                n += 1
                ws = __xlsx_table_to_sheet(t['data'], workbook.create_sheet(str(n)))
            destname = output if output else name + ".%s" % (format)
            workbook.save(destname)
        elif format == "json":
            report = {'filename' : filename, 
            'timestamp' : datetime.datetime.now().isoformat(), 'num_tables' : len(tables),
            'tables' : tables}
            destname = output if output else name + ".%s" % (format)
            f = open(destname, 'w', encoding='utf8')
            json.dump(report, f, ensure_ascii=False, indent=4)
            f.close()

    else:
        for t in tables:
            if lfilter >= len(t):
                continue
            n += 1
            destname = output if output else name + "_%d.%s" % (n, format)
            __store_table(t['data'], destname, format)


def analyze_docx(filename, extract_data=False, strip_space=True):
    """Analyzes docx file and extracts data if requested"""
    tableinfo = []
    document = Document(filename)
    n = 0    
    for table in document.tables:
        n += 1
        info = {}
        info['id'] = n
        info['num_cols'] = len(table.columns)
        info['num_rows'] = len(table.rows)
        info['style'] = table.style.name
        tdata = __extract_table(table, strip_space=strip_space)
        info['data'] = tdata
        tableinfo.append(info)
    return tableinfo


# -*- coding: utf8 -*-

def get_dict_value(adict, key, prefix=None):
    if prefix is None:
        prefix = key.split('.')
    if len(prefix) == 1:
        return adict[prefix[0]]
    else:
        return get_dict_value(adict[prefix[0]], key, prefix=prefix[1:])

def get_dict_value_deep(adict, key, prefix=None, as_array=False, splitter='.'):
    """Used to get value from hierarhic dicts in python with params with dots as splitter"""
    if prefix is None:
        prefix = key.split(splitter)
    if len(prefix) == 1:
        if type(adict) == type({}):
            if prefix[0] not in adict.keys():
                return None
            if as_array:
                return [adict[prefix[0]], ]
            return adict[prefix[0]]
        elif type(adict) == type([]):
            if as_array:
                result = []
                for v in adict:
                    if prefix[0] in v.keys():
                        result.append(v[prefix[0]])
                return result
            else:
                if len(adict) > 0 and prefix[0] in adict[0].keys():
                    return adict[0][prefix[0]]
        return None
    else:
        if type(adict) == type({}):
            if prefix[0] in adict.keys():
                return get_dict_value_deep(adict[prefix[0]], key, prefix=prefix[1:], as_array=as_array)
        elif type(adict) == type([]):
            if as_array:
                result = []
                for v in adict:
                    res = get_dict_value_deep(v[prefix[0]], key, prefix=prefix[1:], as_array=as_array)
                    if res:
                        result.extend(res)
                return result
            else:
                return get_dict_value_deep(adict[0][prefix[0]], key, prefix=prefix[1:], as_array=as_array)
        return None




# -*- coding: utf8 -*-
from ..utils import get_file_type, get_option
import logging
import orjson
import zipfile
from qddate import DateParser
from ..common.scheme import generate_scheme_from_file
from pydantic import BaseModel
from typing import Optional
import duckdb
import pandas as pd
import yaml
import csv
import json
import tempfile
import io
import os
import xxhash
import tqdm
from pyzstd import ZstdFile
from ..ai.perplexity import get_fields_info, get_description



def column_type_parse(column_type):
    is_array = (column_type[-2:] == '[]') 
    if is_array:
        text = column_type[:-2]
    else:
        text = column_type
    if text[:6] == 'STRUCT':        
        atype = text[:6]
    elif text[:4] == 'JSON':
        atype = 'VARCHAR'
    else:
        atype = text
    return [atype, str(is_array)]


def get_schema_key(fields):
    return xxhash.xxh64('|'.join(sorted(fields))).hexdigest()


def duckdb_decompose(filename:str=None, frame:pd.DataFrame=None, filetype:str=None, path:str="*", limit:int=10000000, recursive:bool=True, root:str="", ignore_errors:bool=True):
    """Decomposes file or data frame structure"""  
    text_ignore = ', ignore_errors=true' if ignore_errors else ''
    if filetype in ['csv', 'tsv']:
        read_func = f"read_csv('{filename}'{text_ignore}, sample_size={limit})"
    elif filetype in ['json', 'jsonl']:
        read_func = f"read_json('{filename}'{text_ignore})"
    else:
        read_func = f"'{filename}'"
    if path == '*':
        if filename is not None:
            data = duckdb.sql(f"describe select {path} from {read_func} limit {limit}").fetchall()
        else:
            data = duckdb.sql(f"describe select {path} from frame limit {limit}").fetchall()
    else:
         path_parts = path.split('.')
         query = None
         if len(path_parts) == 1:
            if filename is not None:
                query = f"describe select unnest(\"{path}\", recursive:=true) from {read_func} limit {limit}"
            else:
                query = f"describe select unnest(\"{path}\", recursive:=true) from frame limit {limit}"
         elif len(path_parts) == 2:
            if filename is not None:
                query = f"describe select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from {read_func} limit {limit})"
            else:
                query = f"describe select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from frame limit {limit})"
         elif len(path_parts) == 3:
            if filename is not None:
                query = f"describe select unnest(\"{path_parts[2]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from {read_func} limit {limit}))"
            else:
                query = f"describe select unnest(\"{path_parts[2]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from frame limit {limit}))"                
         elif len(path_parts) == 4:
            if filename is not None:
                query = f"describe select unnest(\"{path_parts[2]}.{path_parts[3]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from {read_func} limit {limit}))"
            else:
                query = f"describe select unnest(\"{path_parts[2]}.{path_parts[3]}\", recursive:=true) from (select unnest(\"{path_parts[1]}\", recursive:=true) from (select unnest(\"{path_parts[0]}\", recursive:=true) from frame limit {limit}))"
         data = duckdb.sql(query).fetchall()
    table = []
    for row in data:        
        item = [row[0] if len(root) ==0 else root+ '.' + row[0], ]
        item.extend(column_type_parse(row[1]))
        table.append(item)
        if recursive and item[1] == 'STRUCT':
            subtable = duckdb_decompose(filename, frame, filetype=filetype, path=row[0] if len(root) == 0 else item[0], limit=limit, recursive=recursive, root=item[0], ignore_errors=ignore_errors)
            for subitem in subtable:
                table.append(subitem)
    return table    


class FieldSchema(BaseModel):
    name: str
    ftype: str
    is_array:bool = False
    description: Optional[str] = None
    sem_type:str = None
    sem_url:str = None


class TableSchema(BaseModel):
    """Table schema"""    
    key: Optional[str] = None
    num_cols: int = -1
    is_flat:bool = True
    id:Optional[str] = None    
    fields:Optional[list[FieldSchema]] = []
    description:Optional[str] = None
    files:Optional[list[str]] = []


MAX_SAMPLE_SIZE = 200
DELIMITED_FILES = ['csv', 'tsv']
DUCKABLE_FILE_TYPES = ['csv', 'jsonl', 'json', 'parquet']  
DUCKABLE_CODECS  = ['zst', 'gzip', 'raw']


def table_from_objects(objects:list, id:str, objects_limit:int, use_pandas:bool=False, filetype='csv', autodoc:bool=False, lang:str='English'):
    """Reconstructs table schema from list of objects"""
    table = TableSchema(id=id)
    table.num_records = len(objects)
    if autodoc:
       f = io.StringIO()
       writer = csv.writer(f)
       writer.writerows(objects[:MAX_SAMPLE_SIZE])
       table.description = get_description(f.getvalue(), language=lang)
    if use_pandas:
        df = pd.DataFrame(objects)
        columns_raw = duckdb_decompose(frame=df, path='*', limit=objects_limit)
    else:
        tfile = tempfile.NamedTemporaryFile(suffix='.' + filetype, mode='w', encoding='utf8', delete=False)
        tfile.close()
        tfile_real = ZstdFile(tfile.name, mode='w', level_or_option=9)
        wrapper = io.TextIOWrapper(tfile_real, encoding='utf8', write_through=True)
        if filetype == 'csv':
            writer = csv.writer(wrapper) 
            writer.writerows(objects[:objects_limit])
        elif filetype == 'jsonl':
            for row in objects[:objects_limit]:            
                wrapper.write(json.dumps(row) + '\n')
        tfile_real.close()
        # Getting structure
        columns_raw = duckdb_decompose(tfile.name, filetype=filetype, path='*', limit=objects_limit)
        os.remove(tfile.name)
    is_flat = True
    table.num_cols = len(columns_raw)

    for column in columns_raw:                          
        field = FieldSchema(name=column[0], ftype=column[1], is_array=column[2])
        table.fields.append(field)
        if field.ftype == 'STRUCT' or field.is_array: is_flat = False                    
        table.is_flat = is_flat                            
    table.num_records = len(objects)    
    return table


def build_schema(filename:str, objects_limit:int=100000):
    fileext = filename.rsplit('.', 1)[-1].lower()
    filetype = fileext
    # Getting total count
    table = TableSchema(id=os.path.basename(filename))               
    # Getting structure
    columns_raw = duckdb_decompose(filename, filetype=filetype, path='*', limit=objects_limit)
    is_flat = True
    table.num_cols = len(columns_raw)
    fieldsnames = []
    for column in columns_raw:                          
        field = FieldSchema(name=column[0], ftype=column[1], is_array=column[2])
        fieldsnames.append(column[0])
        table.fields.append(field)
        if field.ftype == 'STRUCT' or field.is_array: is_flat = False                    
        table.is_flat = is_flat
    table.key = get_schema_key(fieldsnames)
    return table



class Schemer:
    def __init__(self, nodates=True):
        if nodates:
            self.qd = None
        else:
            self.qd = DateParser(generate=True)
        pass

    def extract_schema(self, fromfile, options):
        table = build_schema(fromfile)
        print(yaml.dump(table.model_dump(), Dumper=yaml.Dumper))


    def extract_schema_bulk(self, fromdir, options):
        """Extracts schemes from all data files and writes schema structures"""
        filenames = os.listdir(fromdir)
        files = []
        tables = {}
        for filename in filenames:
            if filename.rsplit('.', 1)[-1] in ['csv', 'json', 'jsonl', 'parquet', 'csv.gz', 'csv.zstd', 'jsonl.zstd']:
                files.append(os.path.join(fromdir, filename))
        mode = options['mode']
        print(f'Found {len(files)} files. Processing mode {mode}')                
        for filename in tqdm.tqdm(files):
            table = build_schema(filename)
            fbase = os.path.basename(filename)
            table.id = table.key
            if mode == 'distinct':
                if table.key not in tables.keys():
                    tables[table.key] = table
                    tables[table.key].files.append(fbase)                
                    print(options)
                    if 'autodoc' in options.keys() and options['autodoc'] and 'lang' in options.keys():                    
                        fields = []
                        for column in table.fields:
                            fields.append(column.name)
                        descriptions = get_fields_info(fields, language=options['lang'])
                        for column in table.fields:
                            if column.name in descriptions.keys(): column.description = descriptions[column.name]                
                else:
                    tables[table.key].files.append(fbase)
            elif mode == 'perfile':
                table.files.append(fbase)                
                if 'autodoc' in options.keys() and options['autodoc'] and 'lang' in options.keys():                    
                    fields = []
                    for column in table.fields:
                        fields.append(column.name)
                    descriptions = get_fields_info(fields, language=options['lang'])
                    for column in table.fields:
                        if column.name in descriptions.keys(): column.description = descriptions[column.name]                
                f = open(os.path.join(options['output'], fbase + '.yaml'), 'w', encoding='utf8')
                f.write(yaml.dump(table.model_dump(), Dumper=yaml.Dumper))
                f.close
        if mode == 'distinct':
            print(f'Total schemas {len(tables)}, files {len(files)}')
        elif mode == 'perfile':
            print(f'Total schemas {len(files)}, files {len(files)}')
        if 'output' in options.keys():
            if mode == 'distinct':
                print(f'Writing schemas')
                for table in tables.values():
                    f = open(os.path.join(options['output'], table.key + '.yaml'), 'w', encoding='utf8')
                    f.write(yaml.dump(table.model_dump(), Dumper=yaml.Dumper))
                    f.close
#            print(yaml.dump(table.model_dump(), Dumper=yaml.Dumper))


    def generate_scheme(self, fromfile, options):
        """Generates cerberus scheme from JSON lines or BSON file"""
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        if f_type not in ['jsonl', 'bson', 'csv']:
            print('Only JSON lines, CSV and BSON (.jsonl, .csv, .bson) files supported now')
            return
        if options['zipfile']:
            z = zipfile.ZipFile(fromfile, mode='r')
            fnames = z.namelist()
            fnames[0]
            if f_type == 'bson':
                infile = z.open(fnames[0], 'rb')
            else:
                infile = z.open(fnames[0], 'r')
        else:
            if f_type == 'bson':
                infile = open(fromfile, 'rb')
            else:
                infile = open(fromfile, 'r', encoding=get_option(options, 'encoding'))

        logging.debug('Start identifying scheme for %s' % (fromfile))
        scheme = generate_scheme_from_file(fileobj=infile, filetype=f_type, delimiter=options['delimiter'], encoding=options['encoding'])
        if options['output']:
            f = open(options['output'], 'w', encoding='utf8')
            f.write(orjson.dumps(scheme, option=orjson.OPT_INDENT_2).decode('utf8'))
            f.close()
        else:
            print(str(orjson.dumps(scheme, option=orjson.OPT_INDENT_2).decode('utf8')))



# -*- coding: utf8 -*-
# import json
import logging
import sys


# from xmlr import xmliter
from ..utils import get_file_type, get_option, strip_dict_fields
from ..common.iterable import IterableData
LINEEND = u'\n'.encode('utf8')

DEFAULT_CHUNK_SIZE = 50

class DataQuery:
    def __init__(self):
        pass


    def query(self, fromfile, options={}):
        """Use mistql to query data"""
        from mistql import query
        f_type = get_file_type(fromfile) if options['format_in'] is None else options['format_in']
        iterable = IterableData(fromfile, options=options)
        to_file = get_option(options, 'output')

        if to_file:
            to_type = get_file_type(to_file)
            if not to_file:
                print('Output file type not supported')
                return
            if to_type == 'bson':
                out = open(to_file, 'wb')
            if to_type == 'jsonl':
                out = open(to_file, 'wb')
            else:
                out = open(to_file, 'w', encoding='utf8')
        else:
            to_type = f_type
            out = sys.stdout
        fields = options['fields'].split(',') if options['fields'] else None
#        writer = DataWriter(out, filetype=to_type, fieldnames=fields)
        if iterable:
            n = 0
            fields = [field.split('.') for field in fields] if fields else None
            for r in iterable.iter():
                n += 1
                if fields:
                    r_selected = strip_dict_fields(r, fields, 0)
                else:
                    r_selected = r
                if options['query'] is not None:
                    res = query(options['query'], r_selected)
                    #                    print(options['filter'], r)
                    if not res:
                        continue
                else:
                    res = r_selected
                print(res)
        else:
            logging.info('File type not supported')
            return
        logging.debug('query: %d records processed' % (n))
        out.close()
